# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

#!/usr/bin/python3
# pylint: disable=invalid-name

"""
This tool scans for a number of common, vulnerable open source components
(openssl, libpng, libxml2, expat and a few others) to let you know if your
system includes common libraries with known vulnerabilities.  It emits a list
of CVE numbers that may be relevant to your binary based on the versions.  It
cannot detect backported fixes.
"""

import argparse
import logging
import os
import platform
import sys
import textwrap
from collections import ChainMap
from typing import Dict

import pkg_resources

from cve_bin_tool.config import ConfigParser
from cve_bin_tool.cve_scanner import CVEScanner
from cve_bin_tool.cvedb import CVEDB, OLD_CACHE_DIR
from cve_bin_tool.error_handler import (
    CVEDataMissing,
    EmptyCache,
    ErrorHandler,
    ErrorMode,
    InsufficientArgs,
    excepthook,
)
from cve_bin_tool.input_engine import InputEngine, TriageData
from cve_bin_tool.log import LOGGER
from cve_bin_tool.merge import MergeReports
from cve_bin_tool.output_engine import OutputEngine
from cve_bin_tool.package_list_parser import PackageListParser
from cve_bin_tool.util import ProductInfo
from cve_bin_tool.version import VERSION
from cve_bin_tool.version_scanner import VersionScanner

sys.excepthook = excepthook  # Always install excepthook for entrypoint module.


class StringToListAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        values = list(map(lambda val: val.strip(), values.split(",")))
        setattr(namespace, self.dest, values)


def main(argv=None):
    """Scan a binary file for certain open source libraries that may have CVEs"""
    argv = argv or sys.argv

    # Reset logger level to info
    LOGGER.setLevel(logging.INFO)

    parser = argparse.ArgumentParser(
        prog="cve-bin-tool",
        description=textwrap.dedent(
            """
            The CVE Binary Tool scans for a number of common, vulnerable open source
            components (openssl, libpng, libxml2, expat and a few others) to let you know
            if a given directory or binary file includes common libraries with known
            vulnerabilities.
            """
        ),
        epilog=textwrap.fill(
            f'Available checkers: {", ".join(VersionScanner.available_checkers())}'
        )
        + "\n\nPlease disclose issues responsibly!",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    input_group = parser.add_argument_group("Input")
    input_group.add_argument(
        "directory", help="directory to scan", nargs="?", default=None
    )
    input_group.add_argument(
        "-e",
        "--exclude",
        action=StringToListAction,
        help="Comma separated Exclude directory path",
        default=None,
    )

    input_group.add_argument(
        "-i",
        "--input-file",
        action="store",
        default="",
        help="provide input filename",
    )
    input_group.add_argument(
        "-C", "--config", action="store", default="", help="provide config file"
    )
    input_group.add_argument(
        "-L", "--package-list", action="store", default="", help="provide package list"
    )

    output_group = parser.add_argument_group("Output")
    output_group.add_argument(
        "-q", "--quiet", action="store_true", help="suppress output"
    )
    output_group.add_argument(
        "-l",
        "--log",
        help="log level (default: info)",
        dest="log_level",
        action="store",
        choices=["debug", "info", "warning", "error", "critical"],
    )
    output_group.add_argument(
        "-o",
        "--output-file",
        action="store",
        help="provide output filename (default: output to stdout)",
    )
    output_group.add_argument(
        "--html-theme",
        action="store",
        help="provide custom theme directory for HTML Report",
    )
    output_group.add_argument(
        "-f",
        "--format",
        action="store",
        choices=["csv", "json", "console", "html", "pdf"],
        help="update output format (default: console)",
    )
    output_group.add_argument(
        "-c",
        "--cvss",
        action="store",
        help="minimum CVSS score (as integer in range 0 to 10) to report (default: 0)",
    )
    output_group.add_argument(
        "-S",
        "--severity",
        action="store",
        choices=["low", "medium", "high", "critical"],
        help="minimum CVE severity to report (default: low)",
    )
    output_group.add_argument(
        "--report",
        action="store_true",
        help="Produces a report even if there are no CVE for the respective output format",
    )
    output_group.add_argument(
        "-a",
        "--append",
        nargs="?",
        const=True,
        help="save output as intermediate report in json format",
    )
    output_group.add_argument(
        "-t",
        "--tag",
        action="store",
        help="add a unique tag to differentiate between multiple intermediate reports",
    )
    parser.add_argument("-V", "--version", action="version", version=VERSION)
    parser.add_argument(
        "-u",
        "--update",
        action="store",
        choices=["now", "daily", "never", "latest"],
        help="update schedule for NVD database (default: daily)",
    )
    parser.add_argument(
        "-x",
        "--extract",
        action="store_true",
        help="autoextract compressed files",
    )
    parser.add_argument(
        "--disable-version-check",
        action="store_true",
        help="skips checking for a new version",
    )
    parser.add_argument(
        "-m",
        "--merge",
        action=StringToListAction,
        help="comma separated intermediate reports path for merging",
    )

    checker_group = parser.add_argument_group("Checkers")
    checker_group.add_argument(
        "-s",
        "--skips",
        dest="skips",
        action=StringToListAction,
        type=str,
        help="comma-separated list of checkers to disable",
    )
    checker_group.add_argument(
        "-r",
        "--runs",
        dest="runs",
        action=StringToListAction,
        type=str,
        help="comma-separated list of checkers to enable",
    )
    defaults = {
        "directory": "",
        "exclude": [],
        "input_file": "",
        "log_level": "info",
        "format": "console",
        "cvss": 0,
        "severity": "low",
        "update": "daily",
        "extract": True,
        "disable_version_check": False,
        "skips": "",
        "runs": "",
        "quiet": False,
        "output_file": "",
        "html_theme": "",
        "report": False,
        "package_list": "",
        "append": False,
        "tag": "",
        "merge": None,
    }

    with ErrorHandler(mode=ErrorMode.NoTrace):
        raw_args = parser.parse_args(argv[1:])
        args = {key: value for key, value in vars(raw_args).items() if value}

    configs = {}
    if args.get("config"):
        conf = ConfigParser(args["config"])
        configs = conf.parse_config()

    args = ChainMap(args, configs, defaults)

    # logging and error related settings
    if args["log_level"]:
        LOGGER.setLevel(args["log_level"].upper())

    if args["quiet"]:
        LOGGER.setLevel(logging.CRITICAL)

    if 0 < LOGGER.level <= 10:
        error_mode = ErrorMode.FullTrace
    elif LOGGER.level >= 50:
        error_mode = ErrorMode.NoTrace
    else:
        error_mode = ErrorMode.TruncTrace

    if platform.system() != "Linux":
        warning_nolinux = """
                          **********************************************
                          Warning: this utility was developed for Linux.
                          You may need to install additional utilities
                          to use it on other operating systems.
                          **********************************************
                          """
        LOGGER.warning(warning_nolinux)

    # CSVScanner related settings
    score = 0
    if args["severity"]:
        # Set minimum CVSS score based on severity
        cvss_score = {"low": 0, "medium": 4, "high": 7, "critical": 9}
        score = cvss_score[args["severity"]]
    if int(args["cvss"]) > 0:
        score = int(args["cvss"])

    merged_reports = None
    if args["merge"]:
        LOGGER.info(
            "You can use -f --format and -o --output-file for saving merged intermediate reports in a file"
        )
        merged_reports = MergeReports(merge_files=args["merge"], score=score)
        if args["input_file"]:
            LOGGER.warning(
                "Ignoring -i --input-file while merging intermediate reports"
            )
            args["input_file"] = None
        merge_cve_scanner = merged_reports.merge_intermediate()

    # Database update related settings
    # Connect to the database
    cvedb_orig = CVEDB(
        version_check=not args["disable_version_check"], error_mode=error_mode
    )

    # if OLD_CACHE_DIR (from cvedb.py) exists, print warning
    if os.path.exists(OLD_CACHE_DIR):
        LOGGER.warning(
            f"Obsolete cache dir {OLD_CACHE_DIR} is no longer needed and can be removed."
        )

    # Clear data if -u now is set
    if args["update"] == "now":
        cvedb_orig.clear_cached_data()

    if args["update"] == "latest":
        cvedb_orig.refresh_cache_and_update_db()

    # update db if needed
    if args["update"] != "never":
        cvedb_orig.get_cvelist_if_stale()
    else:
        LOGGER.warning("Not verifying CVE DB cache")
        cvedb_orig.get_db_update_date()
        if not cvedb_orig.nvd_years():
            with ErrorHandler(mode=error_mode, logger=LOGGER):
                raise EmptyCache(cvedb_orig.cachedir)

    # CVE Database validation
    if not cvedb_orig.check_cve_entries():
        with ErrorHandler(mode=error_mode, logger=LOGGER):
            raise CVEDataMissing("No data in CVE Database")

    cvedb_orig.remove_cache_backup()

    # Input validation
    if (
        not args["directory"]
        and not args["input_file"]
        and not args["package_list"]
        and not args["merge"]
    ):
        parser.print_usage()
        with ErrorHandler(logger=LOGGER, mode=ErrorMode.NoTrace):
            raise InsufficientArgs(
                "Please specify a directory to scan or an input file required"
            )

    # Output validation
    if not args["append"] and args["tag"]:
        LOGGER.warning(
            f"Please specify -a --append to generate intermediate reports while using -t --tag"
        )

    if args["directory"] and not os.path.exists(args["directory"]):
        parser.print_usage()
        with ErrorHandler(logger=LOGGER, mode=ErrorMode.NoTrace):
            raise FileNotFoundError("Directory/File doesn't exist")

    # Checkers related settings
    skips = args["skips"]
    if args["runs"]:
        runs = args["runs"]
        skips = list(
            map(
                lambda checker: checker.name,
                filter(
                    lambda checker: checker.name not in runs,
                    pkg_resources.iter_entry_points("cve_bin_tool.checker"),
                ),
            )
        )

    with CVEScanner(score=score) as cve_scanner:
        triage_data: TriageData
        total_files: int = 0
        parsed_data: Dict[ProductInfo, TriageData] = {}

        # Package List parsing
        if args["package_list"]:
            package_list = PackageListParser(
                args["package_list"], error_mode=error_mode
            )
            parsed_data = package_list.parse_list()
            for product_info, triage_data in parsed_data.items():
                LOGGER.warning(f"{product_info}, {triage_data}")
                cve_scanner.get_cves(product_info, triage_data)

        if args["input_file"]:
            input_engine = InputEngine(
                args["input_file"], logger=LOGGER, error_mode=error_mode
            )
            parsed_data = input_engine.parse_input()
            if not args["directory"]:
                for product_info, triage_data in parsed_data.items():
                    LOGGER.warning(f"{product_info}, {triage_data}")
                    cve_scanner.get_cves(product_info, triage_data)
        if args["directory"]:
            version_scanner = VersionScanner(
                should_extract=args["extract"],
                exclude_folders=args["exclude"],
                error_mode=error_mode,
            )
            version_scanner.remove_skiplist(skips)
            version_scanner.print_checkers()
            for scan_info in version_scanner.recursive_scan(args["directory"]):
                if scan_info:
                    product_info, path = scan_info
                    LOGGER.debug(f"{product_info}: {path}")
                    triage_data = parsed_data.get(product_info, {"default": {}})
                    # Ignore paths from triage_data if we are scanning directory
                    triage_data["paths"] = {path}
                    cve_scanner.get_cves(product_info, triage_data)
            total_files = version_scanner.total_scanned_files

        if args["merge"]:
            cve_scanner = merge_cve_scanner

        LOGGER.info("")
        LOGGER.info("Overall CVE summary: ")
        if args["input_file"] or args["package_list"]:
            LOGGER.info(
                f"There are {cve_scanner.products_with_cve} products with known CVEs detected"
            )
        else:
            LOGGER.info(
                f"There are {cve_scanner.products_with_cve} files with known CVEs detected"
            )
        if cve_scanner.products_with_cve > 0 or args["report"]:
            affected_string = ", ".join(
                map(
                    lambda product_version: "".join(str(product_version)),
                    cve_scanner.affected(),
                )
            )
            LOGGER.info(f"Known CVEs in {affected_string}:")

            # Creates a Object for OutputEngine
            output = OutputEngine(
                all_cve_data=cve_scanner.all_cve_data,
                scanned_dir=args["directory"],
                filename=args["output_file"],
                themes_dir=args["html_theme"],
                time_of_last_update=cvedb_orig.time_of_last_update,
                tag=args["tag"],
                products_with_cve=cve_scanner.products_with_cve,
                products_without_cve=cve_scanner.products_without_cve,
                total_files=total_files,
                is_report=args["report"],
                append=args["append"],
                merge_report=merged_reports,
            )

            if not args["quiet"]:
                output.output_file(args["format"])

        # Use the number of products with known cves as error code
        # as requested by folk planning to automate use of this script.
        # If no cves found, then the program exits cleanly.
        return cve_scanner.products_with_cve


if __name__ == "__main__":
    if os.getenv("NO_EXIT_CVE_NUM"):
        main()
    else:
        sys.exit(main())



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "gcc",
        "version": "9.3.1",
        "version_strings": ["gcc 9.3.1"],
    }
]
package_test_data = [
    {
        "url": "http://mirrors.kernel.org/fedora/releases/33/Everything/x86_64/os/Packages/g/",
        "package_name": "gcc-10.2.1-3.fc33.x86_64.rpm",
        "product": "gcc",
        "version": "10.2.1",
    }
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "xerces-c",
        "version": "3.1.1",
        "version_strings": ["/xerces-c-src_3_1_1/"],
    }
]
package_test_data = [
    {
        "url": "http://mirror.centos.org/centos/7/os/x86_64/Packages/",
        "package_name": "xerces-c-3.1.1-10.el7_7.x86_64.rpm",
        "product": "xerces-c",
        "version": "3.1",
    }
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

import re

from packaging.version import parse

"""
CVE checker for glibc CLI

References:
https://www.gnu.org/software/glibc/
https://www.cvedetails.com/vulnerability-list/vendor_id-72/product_id-767/GNU-Glibc.html


"""
from cve_bin_tool.checkers import Checker


class GlibcChecker(Checker):

    CONTAINS_PATTERNS = [
        r"The following command substitution is needed to make ldd work in SELinux",
        r"environments where the RTLD might not have permission to write to the",
        r"Valid options for the LD_DEBUG environment variable are:",
        r"To direct the debugging output into a file instead of standard output",
        r"Mandatory or optional arguments to long options are also mandatory or optional for any corresponding short options.",
    ]
    FILENAME_PATTERNS = [
        r"ldd",
        r"libc6",
        r"glibc-source",
        r"glibc",
        r"nscd",
        r"libc\.so",
        r"libc-([012](\.[0-9]+){1,2})\.so",
        r"ld-([012](\.[0-9]+){1,2})\.so",
    ]
    VERSION_PATTERNS = [
        r"GNU C Library \(GNU libc\) stable release version ([012](\.[0-9]+){1,2})",
        r"glibc ([012](\.[0-9]+){1,2})",
        r"libc-([012](\.[0-9]+){1,2})\.so\.debug",  # patterns like this aren't ideal (check the end of the file)
        r"ld-([012]\.[0-9]+)\.so\.debug",  # patterns like this aren't ideal
        r"libanl-([012](\.[0-9]+){1,2})\.so\.debug",  # patterns like this aren't ideal
        r"ld-([012](\.[0-9]+){1,2})\.so\.debug",  # patterns like this aren't ideal
    ]
    VENDOR_PRODUCT = [("gnu", "glibc")]

    def get_version(self, lines, filename):

        version_info = super().get_version(lines, filename)
        version_patterns = [r"GLIBC_([0-9]+\.[0-9]+)", r"GLIBC ([0-9]+\.[0-9]+)"]
        if "is_or_contains" in version_info and version_info["version"] == "UNKNOWN":
            # GLIBC* pattern repeats itself a couple of times. Use the latest version value if this pattern exists
            recent = ""
            version_list = []

            for line in lines:
                for pattern in version_patterns:
                    match = re.search(pattern, line)
                    if match:
                        version_list.append(match.group(1).strip())

            if version_list:
                # Find the latest version string pattern
                recent = version_list[0]
                for v in version_list:
                    if parse(recent) < parse(v):
                        recent = v
                version_info["version"] = recent

        return version_info


"""
Using filenames (containing patterns like '.so' etc.) in the binaries as VERSION_PATTERNS aren't ideal.
The reason behind this is that these might depend on who packages the file (like it 
might work on fedora but not on ubuntu)
"""



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {"product": "dovecot", "version": "2.2.11", "version_strings": ["Dovecot v2.2.11"]}
]
package_test_data = [
    {
        "url": "https://kojipkgs.fedoraproject.org/packages/dovecot/2.2.10/1.fc20/x86_64/",
        "package_name": "dovecot-2.2.10-1.fc20.x86_64.rpm",
        "product": "dovecot",
        "version": "2.2.10",
    },
    {
        "url": "http://rpmfind.net/linux/mageia/distrib/5/x86_64/media/core/updates/",
        "package_name": "dovecot-2.2.13-5.1.mga5.x86_64.rpm",
        "product": "dovecot",
        "version": "2.2.13",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for libxml2

References:
http://www.cvedetails.com/vulnerability-list/vendor_id-1962/product_id-3311/Xmlsoft-Libxml2.html

RSS feed: http://www.cvedetails.com/vulnerability-feed.php?vendor_id=1962&product_id=3311&version_id=&orderby=2&cvssscoremin=0
"""
import re

from cve_bin_tool.checkers import Checker


class Xml2Checker(Checker):
    CONTAINS_PATTERNS = [
        r"Internal error, xmlCopyCharMultiByte 0x%X out of bound",
        r"xmlNewElementContent : name != NULL !",
        r"xmlRelaxNG: include %s has a define %s but not the included grammar",
    ]
    FILENAME_PATTERNS = [r"libxml2.so."]
    VERSION_PATTERNS = []
    VENDOR_PRODUCT = [("xmlsoft", "libxml2")]

    @staticmethod
    def guess_xml2_version(lines):
        """Guesses the xml2 version from the file contents"""
        new_guess = ""
        pattern1 = re.compile(r"/libxml2-([0-9]+\.[0-9]+\.[0-9]+)/")
        pattern2 = re.compile(r"\\libxml2-([0-9]+\.[0-9]+\.[0-9]+)\\")
        # fedora 29 string looks like libxml2.so.2.9.8-2.9.8-4.fc29.x86_64.debug
        pattern3 = re.compile(r"libxml2.so.([0-9]+\.[0-9]+\.[0-9]+)")

        for line in lines:
            match = pattern1.search(line)
            if match:
                new_guess2 = match.group(1).strip()
                if len(new_guess2) > len(new_guess):
                    new_guess = new_guess2

            match = pattern2.search(line)
            if match:
                new_guess2 = match.group(1).strip()
                if len(new_guess2) > len(new_guess):
                    new_guess = new_guess2
            if line == "20901":
                new_guess = "2.9.1"
            if line == "20902":
                new_guess = "2.9.2"
            if line == "20903":
                new_guess = "2.9.3"
            if line == "20904":
                new_guess = "2.9.4"

            match = pattern3.search(line)
            if match:
                new_guess2 = match.group(1).strip()
                if len(new_guess2) > len(new_guess):
                    new_guess = new_guess2
        return new_guess

    def get_version(self, lines, filename):
        version_info = super().get_version(lines, filename)
        if version_info:
            version_info["version"] = self.guess_xml2_version(lines)
        return version_info



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

import os
from datetime import datetime
from typing import Union

from jinja2 import Environment, FileSystemLoader, select_autoescape

from cve_bin_tool.merge import MergeReports

from ..util import CVEData


def html_print_mode(
    all_cve_data: CVEData,
    directory: str,
    products_with_cve: int,
    products_without_cve: int,
    total_files: int,
    star_warn: str,
    merge_report: Union[None, MergeReports],
    version: str,
    full_html: bool = True,
) -> str:

    root = os.path.dirname(os.path.abspath(__file__))
    templates_dir = os.path.join(root, "print_mode")
    templates_env = Environment(
        loader=FileSystemLoader(templates_dir),
        autoescape=select_autoescape(
            enabled_extensions=("html"), disabled_extensions=("css,js")
        ),
    )

    temp_showcase = "templates/showcase.html"
    temp_content = "templates/content.html"
    temp_base = "templates/base.html"
    temp_intermediate = "templates/intermediate_content.html"

    showcase = templates_env.get_template(temp_showcase)
    content = templates_env.get_template(temp_content)
    intermediate_content = None
    rendered_report = []
    rendered_report.append(
        showcase.render(
            date=datetime.now().strftime("%d %b %Y"),
            directory=directory,
            products_with_cve=products_with_cve,
            products_without_cve=products_without_cve,
            total_files=total_files,
            version=version,
        )
    )
    if merge_report:
        intermediate_content = templates_env.get_template(temp_intermediate)
        rendered_report.append(
            intermediate_content.render(
                intermediate_data=merge_report.intermediate_cve_data,
            )
        )
    rendered_report.append(
        content.render(
            all_cve_data=all_cve_data, directory=directory, star_warn=star_warn
        )
    )

    rendered_report = "".join(rendered_report)

    if full_html:
        base = templates_env.get_template(temp_base)
        return base.render(content=rendered_report)
    else:
        return rendered_report



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for dpkg

https://www.cvedetails.com/product/18928/?q=Dpkg

"""
from cve_bin_tool.checkers import Checker


class DpkgChecker(Checker):
    CONTAINS_PATTERNS = [
        r"unable to access dpkg status area",
        r"multiple non-coinstallable package instances present; most probably due to an upgrade from an unofficial dpkg",
    ]
    FILENAME_PATTERNS = [r"dpkg"]
    VERSION_PATTERNS = [
        r"dpkg-([0-9]+\.[0-9]+\.[0-9]+)",
        r"dpkg-deb-([0-9]+\.[0-9]+\.[0-9]+)",
        r"dpkg-query-([0-9]+\.[0-9]+\.[0-9]+)",
        r"dpkg-divert-([0-9]+\.[0-9]+\.[0-9]+)",
    ]
    VENDOR_PRODUCT = [("debian", "dpkg")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "qt",
        "version": "4.2.3",
        "version_strings": ["Qt 4.2.3", r"QTest library 4.2.3"],
    }
]
package_test_data = [
    {
        "url": "https://kojipkgs.fedoraproject.org/packages/qt/3.3.8b/8.fc9/i386/",
        "package_name": "qt-3.3.8b-8.fc9.i386.rpm",
        "product": "qt",
        "version": "3.3.8",
    },
    {
        "url": "http://mirror.centos.org/centos/7/os/x86_64/Packages/",
        "package_name": "qt-x11-4.8.7-8.el7.x86_64.rpm",
        "product": "qt",
        "version": "4.8.7",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {"product": "samba", "version": "4.10.2", "version_strings": ["SAMBA_4.10.2"]}
]
package_test_data = [
    {
        "url": "https://kojipkgs.fedoraproject.org/packages/samba/4.10.4/0.fc30/x86_64/",
        "package_name": "samba-4.10.4-0.fc30.x86_64.rpm",
        "product": "samba",
        "version": "4.10.4",
    },
    {
        "url": "http://www.rpmfind.net/linux/fedora/linux/development/rawhide/Everything/aarch64/os/Packages/s/",
        "package_name": "samba-4.13.3-1.fc34.aarch64.rpm",
        "product": "samba",
        "version": "4.13.3",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

"""
CVE-bin-tool CLI tests
"""
import json
import logging
import os
import tempfile
import unittest
from test.utils import (
    CURL_7_20_0_RPM,
    CURL_7_20_0_URL,
    LONG_TESTS,
    TMUX_DEB,
    TMUX_DEB_NAME,
    TempDirTest,
    download_file,
)

import pytest

from cve_bin_tool.cli import main
from cve_bin_tool.cvedb import DISK_LOCATION_DEFAULT
from cve_bin_tool.extractor import Extractor
from cve_bin_tool.version_scanner import VersionScanner


class TestCLI(TempDirTest):
    """Tests the CVE Bin Tool CLI"""

    TEST_PATH = os.path.abspath(os.path.dirname(__file__))

    def setup_method(self):
        download_file(CURL_7_20_0_URL, os.path.join(self.tempdir, CURL_7_20_0_RPM))
        download_file(TMUX_DEB, os.path.join(self.tempdir, TMUX_DEB_NAME))

    def test_extract_curl_7_20_0(self):
        """Scanning curl-7.20.0"""
        assert main(["cve-bin-tool", "-l", "debug", "-x", self.tempdir]) != 0

    def test_binary_curl_7_20_0(self):
        """Extracting from rpm and scanning curl-7.20.0"""
        with Extractor() as ectx:
            extracted_path = ectx.extract(os.path.join(self.tempdir, CURL_7_20_0_RPM))
            assert (
                main(
                    [
                        "cve-bin-tool",
                        "-l",
                        "debug",
                        os.path.join(extracted_path, "usr", "bin", "curl"),
                    ]
                )
                != 0
            )

    def test_no_extraction(self):
        """Test scanner against curl-7.20.0 rpm with extraction turned off"""
        assert main(["cve-bin-tool", os.path.join(self.tempdir, CURL_7_20_0_RPM)]) != 0

    def test_extract_bad_zip_messages(self, caplog):
        """Test that bad zip files are logged as extraction failed, but
        bad exe files produce no such message"""
        BAD_EXE_FILE = os.path.join(
            os.path.join(os.path.abspath(os.path.dirname(__file__)), "assets"),
            "empty-file.exe",
        )
        with caplog.at_level(logging.WARNING):
            main(["cve-bin-tool", BAD_EXE_FILE])
        assert "Failure extracting" not in caplog.text

        BAD_ZIP_FILE = os.path.join(
            os.path.join(os.path.abspath(os.path.dirname(__file__)), "assets"),
            "empty-file.zip",
        )
        with caplog.at_level(logging.WARNING):
            main(["cve-bin-tool", BAD_ZIP_FILE])
        assert "Failure extracting" in caplog.text

    def test_exclude(self, caplog):
        """Test that the exclude paths are not scanned"""
        test_path = os.path.abspath(os.path.dirname(__file__))
        exclude_path = os.path.join(test_path, "assets/")
        checkers = list(VersionScanner().checkers.keys())
        with caplog.at_level(logging.INFO):
            main(["cve-bin-tool", test_path, "-e", ",".join(exclude_path)])
        self.check_exclude_log(caplog, exclude_path, checkers)

    def test_usage(self):
        """Test that the usage returns 0"""
        with pytest.raises(SystemExit) as e:
            main(["cve-bin-tool"])
        assert e.value.args[0] == -6

    def test_invalid_file_or_directory(self):
        """Test behaviour with an invalid file/directory"""
        with pytest.raises(SystemExit) as e:
            main(["cve-bin-tool", "non-existant"])
        assert e.value.args[0] == -3

    def test_invalid_parameter(self):
        """Test that invalid parmeters exit with expected error code.
        ArgParse calls sys.exit(2) for all errors, we've overwritten to -2"""

        # no directory specified
        with pytest.raises(SystemExit) as e:
            main(["cve-bin-tool", "--bad-param"])
        assert e.value.args[0] == -2

        # bad parameter (but good directory)
        with pytest.raises(SystemExit) as e:
            main(["cve-bin-tool", "--bad-param", self.tempdir])
        assert e.value.args[0] == -2

        # worse parameter
        with pytest.raises(SystemExit) as e:
            main(["cve-bin-tool", "--bad-param && cat hi", self.tempdir])
        assert e.value.args[0] == -2

        # bad parameter after directory
        with pytest.raises(SystemExit) as e:
            main(["cve-bin-tool", self.tempdir, "--bad-param;cat hi"])
        assert e.value.args[0] == -2

    @unittest.skipUnless(LONG_TESTS() > 0, "Skipping long tests")
    def test_update_flags(self):
        assert main(["cve-bin-tool", "-x", "-u", "never", self.tempdir]) != 0
        assert main(["cve-bin-tool", "-x", "--update", "daily", self.tempdir]) != 0
        assert main(["cve-bin-tool", "-x", "-u", "now", self.tempdir]) != 0
        with pytest.raises(SystemExit) as e:
            main(["cve-bin-tool", "-u", "whatever", self.tempdir])
        assert e.value.args[0] == -2

    @staticmethod
    def check_exclude_log(caplog, exclude_path, checkers):
        # The final log has all the checkers detected
        final_log = [
            record for record in caplog.records if "NewFound CVEs" in record.message
        ]
        assert len(final_log) == 0, "Checkers from excluded path scanned!!"
        if final_log:
            final_log = final_log[0].message
            for checker in checkers:
                assert checker in final_log, f"found a CVE {checker} in {exclude_path}"

        caplog.clear()

    @staticmethod
    def check_checkers_log(caplog, skip_checkers, include_checkers):
        # The final log has all the checkers detected
        final_log = [
            record for record in caplog.records if "Checkers:" in record.message
        ]
        assert len(final_log) > 0, "Could not find checkers line in log"
        final_log = final_log[0].message
        for checker in skip_checkers:
            assert checker not in final_log, f"found skipped checker {checker}"

        for checker in include_checkers:
            assert checker in final_log, f"could not find expected checker {checker}"
        caplog.clear()

    def test_skips(self, caplog):
        """Tests the skips option"""

        test_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "csv")

        skip_checkers = ["systemd", "xerces", "xml2", "kerberos"]
        include_checkers = ["expat", "libgcrypt", "openssl", "sqlite"]
        with caplog.at_level(logging.INFO):
            main(["cve-bin-tool", test_path, "-s", ",".join(skip_checkers)])
        self.check_checkers_log(caplog, skip_checkers, include_checkers)

        # swap skip_checkers and include_checkers
        include_checkers, skip_checkers = skip_checkers, include_checkers
        with caplog.at_level(logging.INFO):
            main(["cve-bin-tool", test_path, "-s", ",".join(skip_checkers)])
        self.check_checkers_log(caplog, skip_checkers, include_checkers)

    def test_runs(self, caplog):
        test_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "csv")

        runs = ["expat", "libgcrypt", "openssl", "sqlite"]
        skip_checkers = ["systemd", "xerces", "xml2", "kerberos"]
        with caplog.at_level(logging.INFO):
            main(["cve-bin-tool", test_path, "-r", ",".join(runs)])
        self.check_checkers_log(caplog, skip_checkers, runs)

        runs, skip_checkers = skip_checkers, runs
        with caplog.at_level(logging.INFO):
            main(["cve-bin-tool", test_path, "-r", ",".join(runs)])
        self.check_checkers_log(caplog, skip_checkers, runs)

    @unittest.skipUnless(LONG_TESTS() > 0, "Skipping long tests")
    def test_update(self, caplog):
        test_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "csv")

        with caplog.at_level(logging.INFO):
            main(["cve-bin-tool", "-u", "never", test_path])
        assert (
            "cve_bin_tool",
            logging.WARNING,
            "Not verifying CVE DB cache",
        ) in caplog.record_tuples
        caplog.clear()

        with caplog.at_level(logging.INFO):
            main(["cve-bin-tool", "-u", "daily", test_path])
        assert (
            "cve_bin_tool.CVEDB",
            logging.INFO,
            "Using cached CVE data (<24h old). Use -u now to update immediately.",
        ) in caplog.record_tuples or (
            "cve_bin_tool.CVEDB",
            logging.INFO,
            "Updating CVE data. This will take a few minutes.",
        ) in caplog.record_tuples
        caplog.clear()

        with caplog.at_level(logging.INFO):
            main(["cve-bin-tool", "-u", "now", test_path])
        db_path = DISK_LOCATION_DEFAULT
        assert (
            "cve_bin_tool.CVEDB",
            logging.WARNING,
            f"Updating cachedir {db_path}",
        ) in caplog.record_tuples and (
            "cve_bin_tool.CVEDB",
            logging.INFO,
            "Updating CVE data. This will take a few minutes.",
        ) in caplog.record_tuples
        caplog.clear()

        with caplog.at_level(logging.INFO):
            main(["cve-bin-tool", "-u", "latest", test_path])
        assert (
            "cve_bin_tool.CVEDB",
            logging.INFO,
            "Updating CVE data. This will take a few minutes.",
        ) in caplog.record_tuples
        caplog.clear()

    def test_unknown_warning(self, caplog):
        """Test that an "UNKNOWN" file generates a warning"""

        # build the unknown test file in test/binaries
        with tempfile.NamedTemporaryFile(
            "w+b", suffix="png-unknown.out", delete=False
        ) as f:
            signatures = [
                b"\x7f\x45\x4c\x46\x02\x01\x01\x03\n",
                b"Application uses deprecated png_write_init() and should be recompiled",
            ]
            f.writelines(signatures)
            filename = f.name

        # Run against the "unknown" file
        with caplog.at_level(logging.INFO):
            main(["cve-bin-tool", filename])

        # clean up temporary file.
        os.remove(filename)

        warnings = [
            record.message
            for record in caplog.records
            if record.levelname == "WARNING" and record.module == "version_scanner"
        ]
        assert len(warnings) > 0, "Unknown version warning didn't get generated"
        assert f"png was detected with version UNKNOWN in file {filename}" in warnings

    def test_quiet_mode(self, capsys, caplog):
        """Test that an quite mode isn't generating any output"""

        with tempfile.NamedTemporaryFile(
            "w+b", suffix="strong-swan-4.6.3.out", delete=False
        ) as f:
            signatures = [b"\x7f\x45\x4c\x46\x02\x01\x01\x03\n", b"strongSwan 4.6.3"]
            f.writelines(signatures)
            filename = f.name

        main(["cve-bin-tool", "-q", filename, "-u", "now"])
        # clean up temporary file.
        os.remove(filename)

        # Make sure log is empty
        assert not caplog.records

        # Make sure nothing is getting printed on stdout or stderr
        captured = capsys.readouterr()
        assert not (captured.out or captured.err)

    @pytest.mark.parametrize(
        "filename",
        (
            os.path.join(TEST_PATH, "config", "cve_bin_tool_config.toml"),
            os.path.join(TEST_PATH, "config", "cve_bin_tool_config.yaml"),
        ),
    )
    def test_config_file(self, caplog, filename):
        # scan with config file and overwrite output format
        assert main(["cve-bin-tool", "-C", filename, "-l", "info"]) != 0

        # assert only checkers for binutils and curl get to run
        assert (
            "cve_bin_tool.VersionScanner",
            logging.INFO,
            "Checkers: binutils, curl",
        ) in caplog.record_tuples

        # assert only CVEs of curl get reflected. Because other are skipped
        assert (
            "cve_bin_tool",
            logging.INFO,
            "There are 1 products with known CVEs detected",
        ) in caplog.record_tuples

        for record in caplog.record_tuples:
            if record[1] < 20:
                pytest.fail(
                    msg="cli option should override logging level specified in config file"
                )

    @staticmethod
    def check_string_in_file(filename, string_to_find):
        # Check if 'string_to_find' is in file
        fh = open(filename)
        file_contents = fh.readlines()
        fh.close()
        for line in file_contents:
            if string_to_find in line:
                return True
        return False

    def test_severity(self, capsys, caplog):
        # scan with severity setting to ensure only CVEs above severity threshold are reported

        # Check command line parameters - wrong case
        with pytest.raises(SystemExit) as e:
            main(["cve-bin-tool", "-S", "HIGH", self.tempdir])
        assert e.value.args[0] == -2
        # Check command line parameters - wrong option
        with pytest.raises(SystemExit) as e:
            main(["cve-bin-tool", "-S", "ALL", self.tempdir])
        assert e.value.args[0] == -2

        my_test_filename = "sevtest.csv"

        if os.path.exists(my_test_filename):
            os.remove(my_test_filename)
        with caplog.at_level(logging.DEBUG):
            main(
                [
                    "cve-bin-tool",
                    "-x",
                    "-f",
                    "csv",
                    "-o",
                    my_test_filename,
                    "-S",
                    "high",
                    os.path.join(self.tempdir, CURL_7_20_0_RPM),
                ]
            )
        # Verify that no CVEs with a severity of Medium are reported
        assert not self.check_string_in_file(my_test_filename, "MEDIUM")
        # Verify that CVEs with a higher severity are reported
        assert self.check_string_in_file(my_test_filename, "HIGH")
        caplog.clear()
        if os.path.exists(my_test_filename):
            os.remove(my_test_filename)

    def test_CVSS_score(self, capsys, caplog):
        # scan with severity score to ensure only CVEs above score threshold are reported

        my_test_filename = "sevtest.csv"

        # Check command line parameters. Less than 0 result in default behaviour.
        if os.path.exists(my_test_filename):
            os.remove(my_test_filename)
        with caplog.at_level(logging.DEBUG):
            main(
                [
                    "cve-bin-tool",
                    "-x",
                    "-c",
                    "-1",
                    "-f",
                    "csv",
                    "-o",
                    my_test_filename,
                    os.path.join(self.tempdir, CURL_7_20_0_RPM),
                ]
            )
        # Verify that some CVEs with a severity of Medium are reported
        assert self.check_string_in_file(my_test_filename, "MEDIUM")
        caplog.clear()

        # Check command line parameters. >10 results in no CVEs being reported (Maximum CVSS score is 10)
        if os.path.exists(my_test_filename):
            os.remove(my_test_filename)
        with caplog.at_level(logging.DEBUG):
            main(
                [
                    "cve-bin-tool",
                    "-x",
                    "-c",
                    "11",
                    "-f",
                    "csv",
                    "-o",
                    my_test_filename,
                    os.path.join(self.tempdir, CURL_7_20_0_RPM),
                ]
            )
        # Verify that no CVEs are reported (no file is created)
        assert not os.path.exists(my_test_filename)
        caplog.clear()

        with caplog.at_level(logging.DEBUG):
            main(
                [
                    "cve-bin-tool",
                    "-x",
                    "-f",
                    "csv",
                    "-o",
                    my_test_filename,
                    os.path.join(self.tempdir, CURL_7_20_0_RPM),
                ]
            )
        # Verify that CVEs with a severity of Medium are reported
        assert self.check_string_in_file(my_test_filename, "MEDIUM")
        caplog.clear()
        if os.path.exists(my_test_filename):
            os.remove(my_test_filename)

        # Now check subset
        with caplog.at_level(logging.DEBUG):
            main(
                [
                    "cve-bin-tool",
                    "-x",
                    "-c",
                    "7",
                    "-f",
                    "csv",
                    "-o",
                    my_test_filename,
                    os.path.join(self.tempdir, CURL_7_20_0_RPM),
                ]
            )
        # Verify that no CVEs with a severity of Medium are reported
        assert not self.check_string_in_file(my_test_filename, "MEDIUM")
        if os.path.exists(my_test_filename):
            os.remove(my_test_filename)
        caplog.clear()



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for bolt

https://www.cvedetails.com/product/64828/Boltcms-Bolt.html?vendor_id=21391

"""
from cve_bin_tool.checkers import Checker


class BoltChecker(Checker):
    CONTAINS_PATTERNS = [
        r"State of the ForcePower setting of the bolt daemon.",
        r"The generation of the Thunderbolt controller associated",
        r"The maximum generation of any of Thunderbolt controller",
    ]
    FILENAME_PATTERNS = [r"boltd"]
    VERSION_PATTERNS = [r"bolt ([0-9]+\.[0-9]+(\.[0-9])?)"]
    VENDOR_PRODUCT = [("boltcms", "bolt")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Test to check if condensed downloads are committed
"""
import importlib
import subprocess
from test.test_data import __all__ as all_test_name


# Test to check condensed files are committed according to the package test data of checkers
def test_condensed_downloads():
    test_data = list(
        map(lambda x: importlib.import_module(f"test.test_data.{x}"), all_test_name)
    )

    package_names = []
    package_test_data_list = map(lambda x: x.package_test_data, test_data)
    for package_test_data in package_test_data_list:
        for package_data in package_test_data:
            package_names.append(
                "test/condensed-downloads/" + package_data["package_name"] + ".tar.gz"
            )

    condensed_downloads = subprocess.run(
        ["git", "ls-files", "test/condensed-downloads"],
        stdout=subprocess.PIPE,
    )

    condensed_downloads = condensed_downloads.stdout.decode("utf-8")

    assert all(
        item in condensed_downloads for item in package_names
    ), "Condensed downloads are not commited"



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for sqlite

References:
SQLLite version/sha1 data from https://www.sqlite.org/changes.html

CVE list: https://www.cvedetails.com/vulnerability-list/vendor_id-9237/product_id-16355/Sqlite-Sqlite.html

"""
import re
import urllib.error as error
import urllib.request as request

from cve_bin_tool.checkers import Checker
from cve_bin_tool.log import LOGGER
from cve_bin_tool.version_signature import VersionSignatureDb


def get_version_map():
    """Read changelog and get SQLITE_SOURCE_ID to use for versions"""
    version_map = []

    changeurl = "https://www.sqlite.org/changes.html"
    version_pattern = re.compile(r"<h3>\d{4}-\d{2}-\d{2} \((\d+\.\d+[.\d]*)\)</h3>")
    id_patterns = [
        re.compile(r'SQLITE_SOURCE_ID: "([^"]+)"'),
        re.compile(r'"*(\d{4}-\d{2}-\d{2} \d+:\d+:\d+ [\w]+)"*'),
    ]
    try:
        response = request.urlopen(changeurl)
        lines = response.readlines()

        last_version = "UNKNOWN"
        for line_encoded in lines:
            line = line_encoded.decode("UTF-8")

            ver_match = version_pattern.search(line)
            if ver_match:
                last_version = ver_match.group(1)
            for id_pattern in id_patterns:
                id_match = id_pattern.search(line)
                if id_match:
                    version_map.append([last_version, id_match.group(1)])
                    break

    except error.URLError as err:
        LOGGER.error("Could not fetch " + changeurl + ", " + str(err))

    return version_map


class SqliteChecker(Checker):
    CONTAINS_PATTERNS = [
        r"unable to open a temporary database file for storing temporary tables",
        r"json_object() requires an even number of arguments",
        r"ESCAPE expression must be a single character",
        r"SQLite version %s",
    ]
    VENDOR_PRODUCT = [("sqlite", "sqlite")]
    VERSION_PATTERNS = [
        r"Id: SQLite version (\d+\.\d+\.\d+)",
        r"sqlite(\d+)\.debug",
    ]  # patterns like the second one aren't ideal (check the end of the file)
    FILENAME_PATTERNS = [r"sqlite", r"sqlite3"]

    mapdb = VersionSignatureDb("sqlite", get_version_map, 30)
    with mapdb:
        VERSION_MAP = mapdb.get_mapping_data()

    def guess_contains(self, lines):
        """Tries to determine if a file includes sqlite"""
        # since the version strings are super unique here, we can guess the version
        # at the same time

        for line in lines:
            for mapping in self.VERSION_MAP:
                if mapping[1] in line:
                    return True

        # If that fails, find a signature that might indicate presence of sqlite
        return super().guess_contains(lines)

    def get_version(self, lines, filename):
        """returns version information for sqlite as found in a given file.

        The most correct way to do this is to search for the sha1 sums per release.
        Fedora rpms have a simpler SQLite version string.
        If neither of those work, try to at least guess the major version
        """

        version_info = super().get_version(lines, filename)

        for line in lines:
            for mapping in self.VERSION_MAP:
                if mapping[1] in line:
                    # overwrite version with the version found by sha mapping
                    version_info["version"] = mapping[0]

        return version_info


"""
Using filenames (containing patterns like '.so' etc.) in the binaries as VERSION_PATTERNS aren't ideal.
The reason behind this is that these might depend on who packages the file (like it 
might work on fedora but not on ubuntu)
"""



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

#!/usr/bin/python3

"""
These are the customized strings and file classes, by doing this
the tool is able to support other operating systems like Windows
and MacOS.
"""

from cve_bin_tool.async_utils import FileIO, run_coroutine


class Strings:
    # printable characters
    PRINTABLE = set(range(32, 128))
    # add tab to the printable character
    PRINTABLE.add(9)

    def __init__(self, filename=""):
        self.filename = filename
        self.output = [""]

    async def aio_parse(self):
        async with FileIO(self.filename, "rb") as f:
            tmp = []
            async for line in f:
                for char in line:
                    # remove all unprintable characters
                    if char in Strings.PRINTABLE:
                        tmp.append(chr(char))
                    elif tmp:
                        self.output.append("".join(tmp))
                        tmp = []
        return self.output

    def parse(self):
        return run_coroutine(self.aio_parse())



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {"product": "cups", "version": "1.2.4", "version_strings": ["CUPS v1.2.4"]}
]
package_test_data = [
    {
        "url": "https://kojipkgs.fedoraproject.org/packages/cups/1.3.3/1.fc8/x86_64/",
        "package_name": "cups-1.3.3-1.fc8.x86_64.rpm",
        "product": "cups",
        "version": "1.3.3",
    }
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {"product": "openvpn", "version": "2.3.13", "version_strings": ["OpenVPN 2.3.13"]}
]
package_test_data = [
    {
        "url": "https://kojipkgs.fedoraproject.org/packages/openvpn/2.4.1/1.fc25/x86_64/",
        "package_name": "openvpn-2.4.1-1.fc25.x86_64.rpm",
        "product": "openvpn",
        "version": "2.4.1",
    },
    {
        "url": "http://rpmfind.net/linux/mageia/distrib/5/x86_64/media/core/updates/",
        "package_name": "openvpn-2.3.12-1.mga5.x86_64.rpm",
        "product": "openvpn",
        "version": "2.3.12",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

import os

import pytest

from cve_bin_tool.config import ConfigParser
from cve_bin_tool.error_handler import ErrorMode, UnknownConfigType


class TestConfig:
    TEST_PATH = os.path.abspath(os.path.dirname(__file__))
    EXPECTED_CONFIG = {
        "cvss": 0,
        "directory": "test/assets",
        "disable_version_check": False,
        "exclude": [],
        "extract": False,
        "format": "console",
        "input_file": "test/csv/triage.csv",
        "log_level": "debug",
        "output_file": "",
        "quiet": False,
        "runs": ["curl", "binutils"],
        "severity": "low",
        "skips": ["python", "bzip2"],
        "update": "daily",
    }

    def test_non_existent_config(self):
        config_parser = ConfigParser("config.toml", error_mode=ErrorMode.FullTrace)
        with pytest.raises(FileNotFoundError):
            config_parser.parse_config()

    def test_unsupported_config(self):
        config_parser = ConfigParser(
            os.path.join(self.TEST_PATH, "json", "bad.json"),
            error_mode=ErrorMode.FullTrace,
        )
        with pytest.raises(UnknownConfigType):
            config_parser.parse_config()

    @pytest.mark.parametrize(
        "filepath, expected_config",
        (
            (
                os.path.join(TEST_PATH, "config", "cve_bin_tool_config.toml"),
                EXPECTED_CONFIG,
            ),
            (
                os.path.join(TEST_PATH, "config", "cve_bin_tool_config.yaml"),
                EXPECTED_CONFIG,
            ),
        ),
    )
    def test_valid_config(self, filepath, expected_config):
        config_parser = ConfigParser(filepath)
        assert dict(config_parser.parse_config()) == expected_config



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "international_components_for_unicode",
        "version": "60.3",
        "version_strings": ["ICU 60.3"],
    },
    {
        "product": "international_components_for_unicode",
        "version": "60.3",
        "version_strings": ["icu-60.3"],
    },
]
package_test_data = [
    {
        "url": "http://www.rpmfind.net/linux/openmandriva/cooker/repository/x86_64/main/release/",
        "package_name": "icu-68.2-1-omv4002.x86_64.rpm",
        "product": "international_components_for_unicode",
        "version": "68.2",
    },
    {
        "url": "http://mirror.centos.org/centos/8/BaseOS/aarch64/os/Packages/",
        "package_name": "icu-60.3-2.el8_1.aarch64.rpm",
        "product": "international_components_for_unicode",
        "version": "60.3",
    },
    {
        "url": "http://archive.ubuntu.com/ubuntu/pool/main/i/icu/",
        "package_name": "icu-devtools_67.1-4_amd64.deb",
        "product": "international_components_for_unicode",
        "version": "67.1",
    },
    {
        "url": "https://ftp.netbsd.org/pub/pkgsrc/packages/NetBSD/amd64/8.2/All/",
        "package_name": "icu-68.2.tgz",
        "product": "international_components_for_unicode",
        "version": "68.2",
    },
    {
        "url": "https://rpmfind.net/linux/dag/redhat/el4/en/x86_64/dag/RPMS/",
        "package_name": "icu-3.6-1.el4.rf.x86_64.rpm",
        "product": "international_components_for_unicode",
        "version": "3.6",
    },
    # {
    #     "url": "https://rpmfind.net/linux/dag/redhat/el4/en/x86_64/dag/RPMS/",
    #     "package_name": "icu-2.6.2-1.2.el4.rf.x86_64.rpm",
    #     "product": "international_components_for_unicode",
    #     "version": "2.6.2",
    # },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for accountservice

https://www.cvedetails.com/product/63450/?q=accountsservice
https://www.cvedetails.com/product/27451/Canonical-Accountsservice.html?vendor_id=4781
https://www.cvedetails.com/product/48266/Freedesktop-Accountsservice.html?vendor_id=7971
https://www.cvedetails.com/product/63450/Accountsservice-Project-Accountsservice.html?vendor_id=21123

"""
from cve_bin_tool.checkers import Checker


class AccountsserviceChecker(Checker):
    CONTAINS_PATTERNS = []
    FILENAME_PATTERNS = [r"accountsservice"]
    VERSION_PATTERNS = [r"accounts-daemon ([0-9]+\.[0-9]+\.[0-9]+)"]
    VENDOR_PRODUCT = [
        ("canonical", "accountsservice"),
        ("ray_stode", "accountsservice"),
        ("accountsservice_project", "accountsservice"),
        ("freedesktop", "accountsservice"),
    ]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "sqlite",
        "version": "3.30.1",
        "version_strings": [
            "2019-10-10 20:19:45 18db032d058f1436ce3dea84081f4ee5a0f2259ad97301d43c426bc7f3df1b0b",
            "ESCAPE expression must be a single character",
        ],
    },
    {
        "product": "sqlite",
        "version": "3.16.1",
        "version_strings": [
            "2017-01-03 18:27:03 979f04392853b8053817a3eea2fc679947b437fd",
            "ESCAPE expression must be a single character",
        ],
    },
    {
        "product": "sqlite",
        "version": "3.12.2",
        "version_strings": [
            "2016-04-18 17:30:31 92dc59fd5ad66f646666042eb04195e3a61a9e8e",
            "ESCAPE expression must be a single character",
        ],
    },
]
package_test_data = [
    {
        "url": "https://kojipkgs.fedoraproject.org/packages/sqlite/3.16.2/1.fc26/x86_64/",
        "package_name": "sqlite-3.16.2-1.fc26.x86_64.rpm",
        "product": "sqlite",
        "version": "3.16.2",
    },
    {
        "url": "http://rpmfind.net/linux/atrpms/el4-x86_64/atrpms/stable/",
        "package_name": "sqlite-3.1.2-2.99_2.el4.at.i386.rpm",
        "product": "sqlite",
        "version": "3.1.2",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime

from reportlab.graphics.shapes import Drawing, Rect, String
from reportlab.lib import colors
from reportlab.lib.styles import ParagraphStyle as PS
from reportlab.lib.units import cm
from reportlab.platypus import PageBreak, Spacer, TableStyle
from reportlab.platypus.doctemplate import (
    BaseDocTemplate,
    PageTemplate,
    SimpleDocTemplate,
)
from reportlab.platypus.frames import Frame
from reportlab.platypus.paragraph import Paragraph
from reportlab.platypus.tableofcontents import TableOfContents
from reportlab.platypus.tables import Table


class PDFDocTemplate(SimpleDocTemplate):
    def __init__(self, filename, **kw):
        self.allowSplitting = 0
        BaseDocTemplate.__init__(self, filename, **kw)
        template = PageTemplate(
            "normal", [Frame(2.5 * cm, 2.5 * cm, 15 * cm, 25 * cm, id="F1")]
        )
        self.addPageTemplates(template)

    def afterFlowable(self, flowable):
        "Registers TOC entries. and makes each TOC entry clickable"
        if flowable.__class__.__name__ == "Paragraph":
            text = flowable.getPlainText()
            style = flowable.style.name
            if style == "Heading1":
                key = "h1-%s" % self.seq.nextf("heading1")
                self.canv.bookmarkPage(key)
                self.notify("TOCEntry", (0, text, self.page, key))
            if style == "Heading2":
                key = "h2-%s" % self.seq.nextf("heading2")
                self.canv.bookmarkPage(key)
                self.notify("TOCEntry", (1, text, self.page, key))


class ConditionalSpacer(Spacer):
    def wrap(self, availWidth, availHeight):
        height = min(self.height, availHeight - 1e-8)
        return (availWidth, height)


class PDFBuilder:
    h1 = PS(name="Heading1", fontSize=16, fontName="Helvetica-Bold", leading=18)
    h2 = PS(name="Heading2", fontSize=14, fontName="Helvetica-Bold", leading=16)
    h3 = PS(name="Heading3", fontSize=12, fontName="Helvetica", leading=14)

    toc_h1 = PS(
        name="Heading1",
        fontSize=16,
        leftIndent=20,
        firstLineIndent=-20,
        spaceBefore=5,
        fontName="Helvetica-Bold",
        leading=18,
    )
    toc_h2 = PS(
        name="Heading2",
        fontSize=14,
        leftIndent=40,
        firstLineIndent=-20,
        spaceBefore=0,
        fontName="Helvetica-Bold",
        leading=16,
    )

    body = PS(name="body", fontSize=12, fontName="Helvetica", leading=12)

    body_unknown = PS(
        name="body",
        fontSize=12,
        textColor=colors.grey,
        fontName="Helvetica-Bold",
        leading=12,
    )

    body_low = PS(
        name="body",
        fontSize=12,
        textColor=colors.blue,
        fontName="Helvetica-Bold",
        leading=12,
    )

    body_medium = PS(
        name="body",
        fontSize=12,
        textColor=colors.green,
        fontName="Helvetica-Bold",
        leading=12,
    )

    body_high = PS(
        name="body",
        fontSize=12,
        textColor=colors.orange,
        fontName="Helvetica-Bold",
        leading=12,
    )

    body_critical = PS(
        name="body",
        fontSize=12,
        textColor=colors.red,
        fontName="Helvetica-Bold",
        leading=12,
    )

    spacer = ConditionalSpacer(0.25 * cm, 0.25 * cm)

    tblStyle = TableStyle(
        [
            ("INNERGRID", (0, 0), (-1, -1), 0.25, colors.black),
            ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
            ("FONT", (0, 0), (-1, -1), "Helvetica", 12),
            ("FONT", (0, 0), (5, 0), "Helvetica-Bold"),
        ]
    )

    frontPageStyle = TableStyle(
        [
            ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
            ("ALIGN", (0, 0), (-1, -1), "CENTER"),
            ("TOPPADDING", (0, 0), (-1, -1), 20),
        ]
    )

    grey = colors.grey
    blue = colors.blue
    red = colors.red
    green = colors.green
    orange = colors.orange

    cm = cm

    def __init__(self, includeTOC=True):
        self.contents = []
        self.toc_included = includeTOC
        self.toc = TableOfContents()
        self.toc.levelStyles = [self.toc_h1, self.toc_h2]
        self.headingnumber = [0, 0, 0, 0]
        self.table_data = []
        self.note_data = []
        self.table_validation = None

    def _spacer(self):
        self.contents.append(self.spacer)

    def heading(self, level, title):
        self._spacer()
        if level == 1:
            self.headingnumber[level] += 1
            # Reset subheading number
            self.headingnumber[level + 1] = 0
            self.contents.append(
                Paragraph(str(self.headingnumber[level]) + ". " + title, self.h1)
            )
        elif level == 2:
            self.headingnumber[level] += 1
            # Reset subheading number
            self.headingnumber[level + 1] = 0
            self.contents.append(
                Paragraph(
                    str(self.headingnumber[level - 1])
                    + "."
                    + str(self.headingnumber[level])
                    + " "
                    + title,
                    self.h2,
                )
            )
        else:
            print("Ooops.... Level", level)
        self._spacer()

    def paragraph(self, text, style=None):
        if style is None:
            self.contents.append(Paragraph(text, self.body))
        else:
            self.contents.append(Paragraph(text, style))
        self._spacer()

    def createtable(self, ident, header, style, validation=None):
        # Layout is [headings, ....]
        self.table_data = []
        self.table_style = style
        self.table_ident = ident
        # Validation is list of items indicating max length of item in each column.
        # None indicates no limit for column
        if validation is not None:
            # Assume validation for all columns provided
            self.table_validation = validation
        else:
            # No validation specified
            self.table_validation = [None] * len(header)
        if header:
            self.addrow(ident, header)

    def validatedata(self, data):
        i = 0
        newdata = []
        for d in data:
            if self.table_validation[i] is not None:
                # Column size validation
                if len(d) > self.table_validation[i]:
                    if d not in self.note_data:
                        self.note_data.append(d)
                    newdata.append("Note " + str(self.note_data.index(d) + 1))
                else:
                    newdata.append(d)
            else:
                newdata.append(d)
            i += 1
        return newdata

    def addrow(self, ident, data, style=None):
        # Add row to table
        if self.table_ident == ident:
            self.table_data.append(self.validatedata(data))
            if style != None:
                for s in style:
                    # Only take first 4 parameters in each style setting
                    self.table_style.add(s[0], s[1], s[2], s[3])

    def showtable(self, ident, widths=None):
        if self.table_ident == ident:
            self._spacer()
            tbl = Table(self.table_data, colWidths=widths, repeatRows=1)
            tbl.setStyle(self.table_style)
            self.contents.append(tbl)
            self.table_ident = None
            # Optional notes if data in columns truncated
            if len(self.note_data) > 0:
                notes = ["<br/><u>Notes</u>"]
                i = 1
                for d in self.note_data:
                    notes.append(str(i) + ". " + d)
                    i += 1
                self.paragraph("<br/>".join(notes))
                self.note_data = []
            self._spacer()

    def pagebreak(self):
        self.contents.append(PageBreak())

    def _logo(self, text):
        d = Drawing(400, 400)
        d.add(Rect(50, 50, 300, 300, fillColor=colors.green))
        d.add(
            String(
                100,
                200,
                text,
                fontName="Helvetica-Bold",
                fontSize=96,
                fillColor=colors.black,
            )
        )
        d.add(
            String(
                75,
                100,
                "Produced by cve-bin-tool",
                fontName="Helvetica",
                fontSize=24,
                fillColor=colors.black,
            )
        )
        return d

    def front_page(self, pagetitle):
        # Front page
        logo = self._logo("CVE")
        now = datetime.now()
        date_time = now.strftime("%d %B %Y at %H:%M:%S")
        front_page = [
            [logo],
            [Paragraph(pagetitle, self.h1)],
            [Paragraph("Report generated on " + date_time, self.h3)],
        ]
        tbl = Table(front_page, colWidths=10 * cm)
        tbl.setStyle(self.frontPageStyle)
        self.contents.append(tbl)
        self.pagebreak()

    def tableofcontents(self):
        if self.toc_included:
            self.paragraph("Table of Contents")
            self._spacer()
            self.contents.append(self.toc)
            self.pagebreak()

    def pageLayout(self, canvas, doc):
        pageinfo = "CVE Report"
        canvas.saveState()
        canvas.setFont("Helvetica", 10)
        canvas.drawString(10 * cm, 2 * cm, "Page %d" % doc.page)
        canvas.drawString(16 * cm, 28 * cm, "%s" % pageinfo)
        canvas.restoreState()

    def publish(self, filename, includeTOC=True):
        if includeTOC and self.toc_included:
            doc = PDFDocTemplate(filename)
            doc.multiBuild(self.contents, onLaterPages=self.pageLayout)
        else:
            doc = SimpleDocTemplate(filename)
            doc.build(self.contents, onLaterPages=self.pageLayout)



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "python",
        "version": "2.7.0",
        "version_strings": [
            "2.7.0",
            "Internal error in the Python interpreter",
            "Cpython",
            "lib/python2.7",
            "~/anaconda/bin/python",
        ],
    },
    {
        "product": "python",
        "version": "3.7.1",
        "version_strings": [
            "3.7.1",
            "Internal error in the Python interpreter",
            "lib/python3.7",
        ],
    },
]
package_test_data = [
    {
        "url": "https://kojipkgs.fedoraproject.org//packages/python3/3.8.2~rc1/1.fc33/aarch64/",
        "package_name": "python3-3.8.2~rc1-1.fc33.aarch64.rpm",
        "product": "python",
        "version": "3.8.2",
    },
    {
        "url": "https://download-ib01.fedoraproject.org/pub/fedora/linux/releases/33/Everything/aarch64/os/Packages/p/",
        "package_name": "python3-3.9.0-1.fc33.aarch64.rpm",
        "product": "python",
        "version": "3.9.0",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import sqlite3
import time
from datetime import datetime

from cve_bin_tool.cvedb import DISK_LOCATION_DEFAULT


class VersionSignatureDb:
    """Methods for version signature data stored in sqlite"""

    def __init__(self, table_name, mapping_function, duration) -> None:
        """Set location on disk data cache will reside.
        Also sets the table name and refresh duration
        """
        self.table_name = table_name
        self.mapping_function = mapping_function
        self.disk_location = DISK_LOCATION_DEFAULT
        self.duration = duration
        self.conn = None
        self.cursor = None

    @property
    def dbname(self):
        """SQLite datebase file where the data is stored."""
        return os.path.join(self.disk_location, "version_map.db")

    def open(self):
        """Opens connection to sqlite database."""
        self.conn = sqlite3.connect(self.dbname)
        self.cursor = self.conn.cursor()

    def close(self):
        """Closes connection to sqlite database."""
        self.cursor.close()
        self.conn.close()
        self.conn = None
        self.cursor = None

    def __enter__(self):
        """Opens connection to sqlite database."""
        self.open()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Closes connection to sqlite database."""
        self.close()

    def get_mapping_data(self):
        """
        Returns a version map associated with the specified checker. Also takes care of updating
        the data after the specified refresh duration
        """
        self.cursor.execute(
            "CREATE TABLE IF NOT EXISTS {}(version TEXT , sourceId TEXT PRIMARY KEY)".format(
                self.table_name
            )
        )

        self.cursor.execute(
            "CREATE TABLE IF NOT EXISTS {} (datestamp DATETIME PRIMARY KEY)".format(
                "latest_update_" + self.table_name
            )
        )

        update_required: bool = False

        datestamp = self.cursor.execute(
            "SELECT * FROM {}".format("latest_update_" + self.table_name)
        ).fetchone()

        if datestamp and type(datestamp) is int:
            # Updates if the difference between current time and the time of last update is greater than duration

            latest_update = datetime.fromtimestamp(datestamp[0])
            time_elapsed = datetime.now() - latest_update
            if time_elapsed.days >= self.duration:
                update_required = True

        if datestamp is None or update_required:
            # if update is required or database is empty, fetch and insert data into database
            self.cursor.execute(f"DELETE FROM {self.table_name}")
            self.cursor.execute(
                "DELETE FROM {}".format("latest_update_" + self.table_name)
            )
            self.cursor.execute(
                "INSERT INTO {} VALUES (?)".format("latest_update_" + self.table_name),
                (time.time(),),
            )

            for mapping in self.mapping_function():
                self.cursor.execute(
                    "INSERT INTO {} (version, sourceId) VALUES (?, ?)".format(
                        self.table_name
                    ),
                    (mapping[0], mapping[1]),
                )

        data = self.cursor.execute(f"SELECT * FROM {self.table_name}").fetchall()
        self.conn.commit()
        return data



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

import json
import re
from collections import defaultdict
from logging import Logger
from os.path import dirname, getsize, isfile, join
from subprocess import PIPE, run

import distro

from cve_bin_tool.error_handler import (
    EmptyTxtError,
    ErrorHandler,
    ErrorMode,
    InvalidListError,
)
from cve_bin_tool.log import LOGGER
from cve_bin_tool.util import ProductInfo, Remarks

from .vendor_fetch import VendorFetch

ROOT_PATH = join(dirname(__file__), "..")
PYPI_CSV = join(ROOT_PATH, "package_list_parser", "pypi_list.csv")

DEB_DISTROS = ["debian", "pop", "ubuntu"]
RPM_DISTROS = ["centos", "fedora", "opensuse", "rhel"]
SUPPORTED_DISTROS = RPM_DISTROS + DEB_DISTROS


class PackageListParser:
    def __init__(
        self, input_file: str, logger: Logger = None, error_mode=ErrorMode.TruncTrace
    ) -> None:
        self.input_file = input_file
        self.logger = logger or LOGGER.getChild(self.__class__.__name__)
        self.error_mode = error_mode
        self.parsed_data_without_vendor = defaultdict(dict)
        self.parsed_data_with_vendor = defaultdict(dict)
        self.package_names_with_vendor = []
        self.package_names_without_vendor = []

    def parse_list(self):
        input_file = self.input_file
        self.check_file()

        if not input_file.endswith("requirements.txt"):
            if distro.id() not in SUPPORTED_DISTROS:
                LOGGER.warning(
                    f"Package list support only available on {','.join(SUPPORTED_DISTROS)}!"
                )
                return {}

            system_packages = []

            LOGGER.info(f"Scanning {distro.id().capitalize()} package list.")

            if distro.id() in DEB_DISTROS:
                installed_packages = run(
                    [
                        "dpkg-query",
                        "--show",
                        '--showformat={"name": "${binary:Package}", "version": "${Version}"}, ',
                    ],
                    stdout=PIPE,
                )
            elif distro.id() in RPM_DISTROS:
                installed_packages = run(
                    [
                        "rpm",
                        "--query",
                        "--all",
                        "--queryformat",
                        '{"name": "%{NAME}", "version": "%{VERSION}"\\}, ',
                    ],
                    stdout=PIPE,
                )
            installed_packages = json.loads(
                f"[{installed_packages.stdout.decode('utf-8')[0:-2]}]"
            )
            with open(input_file) as req:
                lines = req.readlines()
            for line in lines:
                system_packages.append(re.split("\n", line)[0])

            for installed_package in installed_packages:
                if installed_package["name"] in system_packages:
                    self.package_names_without_vendor.append(installed_package)

        else:
            LOGGER.info("Scanning python package list.")
            txt_package_names = []

            installed_packages_json = run(
                ["pip", "list", "--format", "json"],
                stdout=PIPE,
            )
            installed_packages = json.loads(
                installed_packages_json.stdout.decode("utf-8")
            )

            with open(input_file) as txtfile:
                lines = txtfile.readlines()

                for line in lines:
                    txt_package_names.append(re.split(">|\\[|;|=|\n", line)[0])
                for installed_package in installed_packages:
                    package_name = installed_package["name"].lower()
                    if package_name in txt_package_names:
                        self.package_names_without_vendor.append(installed_package)

        with VendorFetch() as vendor_fetch:
            vendor_package_pairs = vendor_fetch.get_vendor_product_pairs(
                self.package_names_without_vendor
            )

        self.add_vendor(vendor_package_pairs)
        self.parse_data()
        return self.parsed_data_with_vendor

    def add_vendor(self, vendor_package_pairs):
        for vendor_package_pair in vendor_package_pairs:
            for package_name in self.package_names_without_vendor:
                if vendor_package_pair["product"] == package_name["name"].replace(
                    "*", ""
                ):
                    package_name["vendor"] = vendor_package_pair["vendor"] + "*"
                    self.package_names_with_vendor.append(package_name)
                    self.package_names_without_vendor.remove(package_name)
                    break

        for package_name in self.package_names_without_vendor:
            package_name["vendor"] = "UNKNOWN"
            self.package_names_with_vendor.append(package_name)

    def parse_data(self):
        for row in self.package_names_with_vendor:
            product_info = ProductInfo(
                row["vendor"], row["name"].lower(), row["version"]
            )
            self.parsed_data_with_vendor[product_info][
                row.get("cve_number", "").strip() or "default"
            ] = {
                "remarks": Remarks(str(row.get("remarks", "")).strip()),
                "comments": row.get("comments", "").strip(),
                "severity": row.get("severity", "").strip(),
            }
            self.parsed_data_with_vendor[product_info]["paths"] = {""}

    def check_file(self):
        input_file = self.input_file
        error_mode = self.error_mode

        if not isfile(input_file):
            with ErrorHandler(mode=error_mode):
                raise FileNotFoundError(input_file)

        if getsize(input_file) == 0:
            with ErrorHandler(mode=error_mode):
                raise EmptyTxtError(input_file)

        if not input_file.endswith("requirements.txt"):
            if distro.id() in DEB_DISTROS:
                # Simulate installation on Debian based system using apt-get to check if the file is valid
                output = run(
                    [f"xargs", "-a", input_file, "apt-get", "install", "-s"],
                    stderr=PIPE,
                    stdout=PIPE,
                )

                if output.returncode != 0:
                    with ErrorHandler(mode=error_mode):
                        raise InvalidListError(
                            f"Invalid Package list\n{output.stderr.decode('utf-8')}"
                        )
            elif distro.id() in RPM_DISTROS:
                output = run(
                    [f"xargs", "-a", input_file, "rpm", "-qi"],
                    stderr=PIPE,
                    stdout=PIPE,
                )

                not_installed_packages = re.findall(
                    r"package (.+) is not installed", output.stdout.decode("utf-8")
                )
                if not_installed_packages:
                    with ErrorHandler(mode=error_mode):
                        raise InvalidListError(
                            f"The packages {','.join(not_installed_packages)} seems to be not installed.\nIt is either an invalid package or not installed.\nUse `sudo yum install $(cat package-list)` to install all packages"
                        )

            else:
                # TODO: Replace below error handling with a proper pip install dry run
                # See: https://github.com/pypa/pip/issues/53
                with ErrorHandler(mode=error_mode):
                    raise InvalidListError("Invalid Python Package list")



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for libpng

References:
http://www.cvedetails.com/vulnerability-list/vendor_id-7294/Libpng.html

RSS feed: http://www.cvedetails.com/vulnerability-feed.php?vendor_id=7294&product_id=0&version_id=&orderby=3&cvssscoremin=0
"""
from cve_bin_tool.checkers import Checker


class PngChecker(Checker):
    CONTAINS_PATTERNS = [
        r"libpng error: %s, offset=%d",
        r"Application uses deprecated png_write_init\(\) and should be recompiled",
        r"libpng version ",
    ]
    FILENAME_PATTERNS = [r"libpng.so.", r"libpng16.so."]
    VERSION_PATTERNS = [r"libpng version ([0-9]+\.[0-9]+\.[0-9]+) -"]
    VENDOR_PRODUCT = [("libpng", "libpng")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import re
import subprocess
import sys
import textwrap

from rich import print as rprint
from rich.console import Console

from cve_bin_tool.cvedb import CVEDB, DBNAME, DISK_LOCATION_DEFAULT
from cve_bin_tool.error_handler import (
    ErrorHandler,
    ErrorMode,
    InsufficientArgs,
    UnknownArchiveType,
)
from cve_bin_tool.extractor import Extractor
from cve_bin_tool.file import is_binary
from cve_bin_tool.log import LOGGER
from cve_bin_tool.strings import Strings
from cve_bin_tool.util import DirWalk, inpath
from cve_bin_tool.version_scanner import (  # using it just for clean_file_path
    VersionScanner,
)


class HelperScript:
    """Helps contributors who want to write a new cve-bin-tool checker find common filenames, version strings, and other necessary data for building a binary checker"""

    CONSOLE = Console()

    def __init__(self, filename, product_name=None):
        self.filename = filename
        self.extractor = Extractor()
        self.logger = LOGGER.getChild(self.__class__.__name__)
        self.product, self.version = self.parse_filename(filename)
        if product_name:
            self.product = product_name

        # for setting the database
        self.connection = None
        self.dbpath = os.path.join(DISK_LOCATION_DEFAULT, DBNAME)

        # for extraction
        self.walker = DirWalk().walk
        self.file_stack = []  # stores extracted filepaths

        # for output (would use in future)
        self.contain_patterns = []
        self.filename_pattern = []
        self.version_pattern = []
        self.vendor_product = self.find_vendor_product()

    def extract_and_parse_file(self, filename):
        """extracts and parses the file for common patterns, version strings and common filename patterns"""

        with self.extractor as ectx:
            if ectx.can_extract(filename):
                binary_string_list = []
                for filepath in self.walker([ectx.extract(filename)]):
                    clean_path = VersionScanner.clean_file_path(filepath)
                    self.file_stack.append(f"{clean_path}")
                    # self.logger.warning(f"scanning binaries from {clean_path}")

                    # see if the file is ELF binary file and parse for strings
                    if self.is_executable(filepath):
                        # self.logger.warning(f"{clean_path} <--- this is an ELF binary")
                        string_list = [i.strip() for i in self.parse_strings(filepath)]
                        matches = self.search_pattern(string_list)

                        # searching for version strings in the found matches
                        version_string = self.search_version_string(matches)
                        self.version_pattern += version_string

                        # if version string is found in file, append it to filename_pattern
                        if version_string:
                            if sys.platform == "win32":
                                self.filename_pattern.append(filepath.split("\\")[-1])
                            else:
                                self.filename_pattern.append(filepath.split("/")[-1])
                            binary_string_list += matches

                            for i in matches:
                                if ("/" not in i and "!" not in i) and len(i) > 40:
                                    self.contain_patterns.append(i)

                        # self.logger.debug(f"{self.filename_pattern}")

                # to resolve case when there are no strings common with product_name in them
                if self.contain_patterns:
                    return self.contain_patterns
                else:
                    return binary_string_list

    def is_executable(self, filename):
        """checks if given file is executable/ELF binary file"""

        if inpath("file"):
            # using system file if available (for performance reasons)
            o = subprocess.check_output(["file", filename])
            o = o.decode(sys.stdout.encoding)

            if (
                ("LSB " not in o)
                and ("LSB shared" not in o)
                and ("LSB executable" not in o)
                and ("PE32 executable" not in o)
                and ("PE32+ executable" not in o)
                and ("Mach-O" not in o)
            ):
                return False
        # else using python implementation of file
        elif not is_binary(filename):
            return False

        return True

    def parse_strings(self, filename):
        """parse binary file's strings"""

        # self.logger.debug(f"running for strings over {filepath}")
        if inpath("strings"):
            # use "strings" on system if available (for performance)
            o = subprocess.check_output(["strings", filename])
            lines = o.decode("utf-8").splitlines()
        else:
            # Otherwise, use python implementation
            s = Strings(filename)
            lines = s.parse()
        return lines

    def search_pattern(self, string_list):
        """find strings for CONTAIN_PATTERNS with product_name in them"""

        pattern = rf"{self.product}"
        matched = [
            i.strip() for i in string_list if re.search(pattern, i, re.IGNORECASE)
        ]
        return matched

    def search_version_string(self, matched_list):
        """finds version strings from matched list"""

        # pattern = rf"{self.product}(.*){self.version}"
        # ^ this does not work for debian packages
        pattern = rf"{self.product}(.*)([0-9]+[.-][0-9]+([.-][0-9]+)?)"
        # this matches patterns like:
        # product1.2.3
        # product 1.2.3
        # product-1.2.3
        # product.1.2.3
        # product version 1.2.3
        # product v1.2.3(1)

        version_strings = [
            i
            for i in matched_list
            if re.search(pattern, i, re.IGNORECASE)
            if not i.endswith(
                ".debug"
            )  # removes .debug, so, this does not gets printed
        ]
        return version_strings

    def parse_filename(self, filename):
        """
        returns package_name/product_name from package_filename of types .rpm, .deb, etc.
        Example: package_filename = openssh-client_8.4p1-5ubuntu1_amd64.deb
            here, package_name = openssh-client
        """

        # resolving directory names
        if sys.platform == "win32":
            filename = filename.split("\\")[-1]
        else:
            filename = filename.split("/")[-1]

        # if extractable, then parsing for different types of files accordingly
        if self.extractor.can_extract(filename):
            if filename.endswith(".tar.xz"):
                product_name = filename.rsplit("-", 3)[0]
                version_number = filename.rsplit("-", 3)[1]
                # example: libarchive-3.5.1-1-aarch64.pkg.tar.xz
            elif filename.endswith(".deb") or filename.endswith(".ipk"):
                product_name = filename.rsplit("_")[0]
                version_number = filename.rsplit("_")[1]
                # example: varnish_6.4.0-3_amd64.deb
            else:
                product_name = filename.rsplit("-", 2)[0]
                version_number = filename.rsplit("-", 2)[1]

            self.logger.debug(
                f"The product and version are '{product_name}', '{version_number}'"
            )
            return product_name, version_number
        else:
            ### raise error for unknown archive types
            with ErrorHandler(mode=ErrorMode.NoTrace, logger=self.logger):
                raise UnknownArchiveType(filename)

    def find_vendor_product(self):
        """find vendor-product pairs from database"""

        # self.logger.debug(f"product_name='{self.product}'")

        CVEDB.db_open(self)
        cursor = self.connection.cursor()

        # finding out all distinct (vendor, product) pairs with the help of product_name
        query = f"""
            SELECT distinct vendor, product FROM cve_range
            WHERE product=(:product);
        """

        cursor.execute(query, {"product": self.product})
        data = cursor.fetchall()

        # checking if (vendor, product) was found in the database
        if data:
            # warning the user to select the vendor-product pairs manually if multiple pairs are found
            if len(data) != 1:
                self.logger.warning(
                    textwrap.dedent(
                        f"""
                            ============================================
                            Multiple ("vendor", "product") pairs found.
                            Please manually select the appropriate pair.
                            ============================================
                        """
                    )
                )
            return data  # [('vendor', 'product')]
        else:
            if self.product:
                # removing numeric characters from the product_name
                if any(char.isdigit() for char in self.product):
                    # self.logger.debug(f"removing digits from product_name.")
                    self.product = "".join(
                        filter(lambda x: not x.isdigit(), self.product)
                    )
                    return self.find_vendor_product()
                else:
                    ### raise error and ask for product_name
                    self.logger.warning(
                        textwrap.dedent(
                            f"""
                                =================================================================
                                No match was found for "{self.product}" in database. 
                                Please check your file or try specifying the "product_name" also.
                                =================================================================
                            """
                        )
                    )
                    return []

        CVEDB.db_close(self)

    def output(self):
        """display beautiful output for Helper-Script"""

        self.CONSOLE.rule(f"[bold dark_magenta]{self.product.capitalize()}Checker")

        # output: long human readable strings
        print("CONTAIN_PATTERNS = [")
        for common_strings in sorted(self.contain_patterns):
            if ".debug" in common_strings:
                rprint(
                    f'\t[red]r"{common_strings}"[/] <--- not recommended to use these form of strings'
                )
                continue  # without this, the else statement was getting printed ;-;
            if ".so" in common_strings:
                rprint(
                    f'\t[red]r"{common_strings}"[/] <--- not recommended to use these form of strings'
                )
            else:
                rprint(f'\t[green]r"{common_strings}"[/],')
        print("]")

        """
        Using filenames (containing patterns like '.so' etc.) in the binaries as VERSION_PATTERNS aren't ideal.
        The reason behind this is that these might depend on who packages the file (like it 
        might work on fedora but not on ubuntu)
        """

        # output: filenames, that we search for binary strings
        print("FILENAME_PATTERNS = [")
        for filename in self.filename_pattern:
            if self.product == filename:
                rprint(
                    f'\t[cyan]r"{filename}"[/], <--- this is a really common filename pattern'
                )
            elif self.product in filename:
                if ".so" in filename:
                    rprint(f'\t[green]r"{filename}"[/],')
                else:
                    rprint(
                        f'\t[bright_green]r"{filename}"[/], <--- you could just use "{self.product}" to match this file'
                    )  # to single-handedly match filenames of type varnishd, varnishlog, varnishtop, etc.
            else:
                rprint(f'\t[green]r"{filename}"[/],')
        print("]")

        # output: version-strings
        print("VERSION_PATTERNS = [")
        for version_string in self.version_pattern:
            rprint(f'\t[green]r"{version_string}"[/],')
        print("]")

        # output: vendor-product pair
        print("VENDOR_PRODUCT = ", end="")
        rprint(self.vendor_product)

        self.CONSOLE.rule()


def main(filenames, product_name=None):
    # finds common strings across multiple files for CONTAIN_PATTERNS
    hs = HelperScript(filenames[1], product_name=product_name)
    binary_string_list_1 = hs.extract_and_parse_file(filenames[1])
    binary_string_list_2 = []

    for filename in filenames[2:]:
        hs = HelperScript(filename)
        binary_string_list_2 = hs.extract_and_parse_file(filename)

        if binary_string_list_2:
            binary_string_list_1 = list(
                set(binary_string_list_1).intersection(set(binary_string_list_2))
            )

            hs.contain_patterns = binary_string_list_1
            binary_string_list_2 = []

    hs.output()


if __name__ == "__main__":
    # accepting arguments
    args = sys.argv

    ### raise error if no file given
    if len(args) == 1:
        with ErrorHandler(mode=ErrorMode.NoTrace, logger=LOGGER):
            raise InsufficientArgs("Atleast 1 input file is required")

    # when product_name is given as last argument
    # product_name must only contain alphabets
    if args[-1].isalpha() and len(args) == 3:
        filenames, product_name = args[:-1], args[-1]
        main(filenames, product_name)
    else:
        filenames = args[:]
        main(filenames)



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

# !/usr/bin/env python3
"""
CVE checker for Gstreamer

References:
https://www.cvedetails.com/vulnerability-list/vendor_id-9481/Gstreamer.html
"""
from cve_bin_tool.checkers import Checker


class GstreamerChecker(Checker):
    CONTAINS_PATTERNS = [
        r"http://bugzilla.gnome.org/enter_bug.cgi?product=GStreamer",
    ]
    FILENAME_PATTERNS = [r"gstreamer"]
    VERSION_PATTERNS = [r"libgstreamer-((\d+\.)+\d+)"]
    VENDOR_PRODUCT = [("gstreamer_project", "gstreamer")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "libjpeg-turbo",
        "version": "2.0.1",
        "version_strings": [
            "LIBJPEGTURBO_2.0.1",
            "Invalid JPEG file structure: two SOF markers",
        ],
    }
]
package_test_data = [
    {
        "url": "https://archives.fedoraproject.org/pub/archive/fedora/linux/releases/30/Everything/x86_64/os/Packages/l/",
        "package_name": "libjpeg-turbo-2.0.2-1.fc30.x86_64.rpm",
        "product": "libjpeg-turbo",
        "version": "2.0.2",
    }
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

"""
CVE checker for gpgme

https://www.cvedetails.com/product/94121/?q=Gpgme
https://www.cvedetails.com/product/10513/GNU-Gpgme.html?vendor_id=72

"""
from cve_bin_tool.checkers import Checker


class GpgmeChecker(Checker):
    CONTAINS_PATTERNS = []
    FILENAME_PATTERNS = [r"gpgme"]
    VERSION_PATTERNS = [
        r"This is GPGME ([0-9]+\.[0-9]+\.[0-9]+) \- The GnuPG Made Easy library",
        r"GPGME-Tool ([0-9]+\.[0-9]+\.[0-9]+) ready",
    ]
    VENDOR_PRODUCT = [("gnupg", "gpgme"), ("gnu", "gpgme")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "pspp",
        "version": "1.2.0",
        "version_strings": ["libpspp-1.2.0"],
    },
    {
        "product": "pspp",
        "version": "1.4.0",
        "version_strings": ["libpspp-1.4.0"],
    },
]
package_test_data = [
    {
        "url": "https://www.rpmfind.net/linux/opensuse/distribution/leap/15.1/repo/oss/x86_64/",
        "package_name": "pspp-1.2.0-lp151.2.1.x86_64.rpm",
        "product": "pspp",
        "version": "1.2.0",
    },
    {
        "url": "https://download-ib01.fedoraproject.org/pub/fedora/linux/releases/33/Everything/aarch64/os/Packages/p/",
        "package_name": "pspp-1.4.0-1.fc33.aarch64.rpm",
        "product": "pspp",
        "version": "1.4.0",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "lighttpd",
        "version": "1.4.30",
        "version_strings": ["Invalid fds at startup with lighttpd", "lighttpd/1.4.30"],
    }
]
package_test_data = [
    {
        "url": "http://mirrors.kernel.org/fedora/releases/33/Everything/x86_64/os/Packages/l/",
        "package_name": "lighttpd-1.4.55-4.fc33.x86_64.rpm",
        "product": "lighttpd",
        "version": "1.4.55",
    },
    {
        "url": "https://ftp.lysator.liu.se/pub/opensuse/distribution/leap/15.1/repo/oss/x86_64/",
        "package_name": "lighttpd-1.4.49-lp151.2.3.x86_64.rpm",
        "product": "lighttpd",
        "version": "1.4.49",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for subversion CLI
References:
https://subversion.apache.org/
https://www.cvedetails.com/vulnerability-list/vendor_id-2120/product_id-3608/Subversion-Subversion.html
"""
from cve_bin_tool.checkers import Checker


class SubversionChecker(Checker):

    CONTAINS_PATTERNS = [
        r"Working copy locked; if no other Subversion client is currently using the working copy, try running 'svn cleanup' without the --remove",
        r"Working copy locked; try running 'svn cleanup' on the root of the working copy ('%s') instead.",
    ]
    FILENAME_PATTERNS = [
        r"subversion",
        r"svn",
        r"libapache2-svn",
        r"libsvn-dev",
        r"svnadmin",
        r"svnauthz",
        r"svnauthz-validate",
        r"svnbench",
        r"svndumpfilter",
        r"svnfsfs",
        r"svnlook",
        r"svnmucc",
        r"svnrdump",
        r"svnserve",
        r"svnsync",
        r"svnversion",
    ]
    VERSION_PATTERNS = [
        r"subversion-([0-9]+\.[0-9]+\.[0-9]+)",
    ]
    VENDOR_PRODUCT = [("subversion", "subversion"), ("apache", "subversion")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

__all__ = [
    "accountsservice",
    "avahi",
    "bash",
    "bind",
    "binutils",
    "bolt",
    "bubblewrap",
    "busybox",
    "bzip2",
    "cryptsetup",
    "cups",
    "curl",
    "dnsmasq",
    "dovecot",
    "dpkg",
    "enscript",
    "expat",
    "ffmpeg",
    "freeradius",
    "ftp",
    "gcc",
    "gimp",
    "glibc",
    "gnupg",
    "gnutls",
    "gpgme",
    "gstreamer",
    "haproxy",
    "hostapd",
    "icecast",
    "international_components_for_unicode",
    "irssi",
    "kerberos",
    "kerberos_5",
    "libarchive",
    "libcurl",
    "libdb",
    "libgcrypt",
    "libjpeg-turbo",
    "libsndfile",
    "libsoup",
    "libssh2",
    "libvirt",
    "libxslt",
    "logrotate",
    "mariadb",
    "memcached",
    "mysql",
    "netpbm",
    "nss",
    "tiff",
    "lighttpd",
    "ncurses",
    "nessus",
    "nginx",
    "node",
    "ntp",
    "openafs",
    "openjpeg",
    "openldap",
    "openssh",
    "openssl",
    "openswan",
    "openvpn",
    "p7zip",
    "png",
    "polarssl",
    "postgresql",
    "pspp",
    "python",
    "qt",
    "radare2",
    "rsyslog",
    "samba",
    "sqlite",
    "strongswan",
    "subversion",
    "syslog-ng",
    "systemd",
    "tcpdump",
    "varnish",
    "webkitgtk",
    "wireshark",
    "wpa_supplicant",
    "xerces",
    "xml2",
    "zlib",
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import sys
from io import StringIO

from setuptools.dist import Distribution


def IS_DEVELOP() -> bool:
    return any(
        list(
            map(
                os.path.isfile,
                list(
                    map(
                        lambda syspath: os.path.join(syspath, "cve-bin-tool.egg-link"),
                        sys.path,
                    )
                ),
            )
        )
    )


def update_egg() -> None:
    with StringIO() as f:
        cwd = os.getcwd()
        os.chdir(os.path.join(os.path.dirname(__file__), ".."))
        sys.stdout = f
        sys.stderr = f
        dist = Distribution(
            dict(
                script_name="setup.py",
                script_args=["egg_info"],
                name="cve-bin-tool",
                entry_points={
                    "cve_bin_tool.checker": [
                        "{} = cve_bin_tool.checkers.{}:{}".format(
                            filename.replace(".py", ""),
                            filename.replace(".py", ""),
                            "".join(
                                (filename.replace(".py", "") + " checker")
                                .replace("_", " ")
                                .title()
                                .split()
                            ),
                        )
                        for filename in os.listdir(
                            os.path.join(
                                os.path.abspath(os.path.dirname(__file__)),
                                "checkers",
                            )
                        )
                        if filename.endswith(".py") and "__init__" not in filename
                    ],
                },
            )
        )
        dist.parse_command_line()
        dist.run_commands()
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        os.chdir(cwd)


if __name__ == "__main__":
    update_egg()



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for zlib
--------
References:
http://www.cvedetails.com/vulnerability-list/vendor_id-72/product_id-1820/GNU-Zlib.html
https://zlib.net/ChangeLog.txt

RSS feed: http://www.cvedetails.com/vulnerability-feed.php?vendor_id=72&product_id=1820&version_id=&orderby=2&cvssscoremin=0
"""
from string import printable

from cve_bin_tool.checkers import Checker


class ZlibChecker(Checker):
    CONTAINS_PATTERNS = [
        rf"deflate[{printable}]*Copyright 1995-2005 Jean-loup Gailly",
        rf"Copyright 1995-2005 Mark Adler[{printable}]*inflate",
        r"too many length or distance symbols",
        r"Copyright 1995-2017 Jean-loup Gailly and Mark Adler",
        r"Copyright 1995-2017 Mark Adler",
    ]
    FILENAME_PATTERNS = [r"libz.so."]
    VERSION_PATTERNS = [
        r"deflate ([01]+\.[0-9]+\.[0-9]+) ",
        r"inflate ([01]+\.[0-9]+\.[0-9]+) ",
        r"libz.so.([01]+\.[0-9]+\.[0-9]+)",  # patterns like this aren't ideal
    ]
    VENDOR_PRODUCT = [("gnu", "zlib")]


"""
Using filenames (containing patterns like '.so' etc.) in the binaries as VERSION_PATTERNS aren't ideal.
The reason behind this is that these might depend on who packages the file (like it 
might work on fedora but not on ubuntu)
"""



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {"product": "netpbm", "version": "10.35.32", "version_strings": ["Netpbm 10.35.32"]}
]
package_test_data = [
    {
        "url": "https://kojipkgs.fedoraproject.org/packages/netpbm/10.35.46/1.fc10/x86_64/",
        "package_name": "netpbm-10.35.46-1.fc10.x86_64.rpm",
        "product": "netpbm",
        "version": "10.35.46",
    },
    {
        "url": "http://vault.centos.org/4.9/os/x86_64/CentOS/RPMS/",
        "package_name": "netpbm-10.35.58-6.el4.x86_64.rpm",
        "product": "netpbm",
        "version": "10.35.58",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for irssi

https://www.cvedetails.com/product/2131/Irssi-Irssi.html?vendor_id=1229

"""
from cve_bin_tool.checkers import Checker


class IrssiChecker(Checker):
    CONTAINS_PATTERNS = []
    FILENAME_PATTERNS = [r"irssi"]
    VERSION_PATTERNS = [r"irssi ([0-9]+\.[0-9]+\.[0-9]+)"]
    VENDOR_PRODUCT = [("irssi", "irssi")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for enscript

https://www.cvedetails.com/product/1800/?q=Enscript

"""
from cve_bin_tool.checkers import Checker


class EnscriptChecker(Checker):
    CONTAINS_PATTERNS = [
        r"set the PostScript language level that enscript",
        r"or set the environment variable `ENSCRIPT_LIBRARY' to point to your library directory.",
    ]
    FILENAME_PATTERNS = [r"enscript"]
    VERSION_PATTERNS = [r"GNU Enscript ([0-9]+\.[0-9]+\.[0-9]+)"]
    VENDOR_PRODUCT = [("gnu", "enscript")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for node.js

References:
http://www.cvedetails.com/vulnerability-list/vendor_id-12113/Nodejs.html

RSS feed: http://www.cvedetails.com/vulnerability-feed.php?vendor_id=12113&product_id=0&version_id=&orderby=3&cvssscoremin=0
"""
from cve_bin_tool.checkers import Checker


class NodeChecker(Checker):
    CONTAINS_PATTERNS = [
        r"https://nodejs.org/download/release/v",
        r"(0) == (uv_async_init(uv_default_loop(), &dispatch_debug_messages_async, DispatchDebugMessagesAsyncCallback))",
        r"Documentation can be found at http://nodejs.org/",
    ]
    FILENAME_PATTERNS = [r"bin/node"]
    VERSION_PATTERNS = [
        r"https\:\/\/nodejs.org\/download\/release\/v([0-9]+\.[0-9]+\.[0-9]+)\/",
        r"node v([0-9]+\.[0-9]+\.[0-9]+)",
    ]
    VENDOR_PRODUCT = [("nodejs", "node.js")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

import json
import os
from collections import defaultdict
from datetime import datetime
from logging import Logger
from typing import Dict, List

from cve_bin_tool.cve_scanner import CVEScanner

from .cvedb import DISK_LOCATION_DEFAULT
from .error_handler import (
    ErrorHandler,
    ErrorMode,
    InvalidIntermediateJsonError,
    InvalidJsonError,
    MissingFieldsError,
)
from .input_engine import TriageData
from .log import LOGGER
from .util import DirWalk, ProductInfo, Remarks

REQUIRED_INTERMEDIATE_METADATA = {
    "scanned_dir",
    "total_files",
    "products_without_cve",
    "products_with_cve",
    "tag",
    "timestamp",
}


class MergeReports:
    def __init__(
        self,
        merge_files: List[str],
        logger: Logger = None,
        error_mode=ErrorMode.TruncTrace,
        cache_dir=DISK_LOCATION_DEFAULT,
        score=0,
    ):
        self.logger = logger or LOGGER.getChild(self.__class__.__name__)
        self.merge_files = merge_files
        self.intermediate_cve_data = []
        self.all_cve_data = []
        self.file_stack = []
        self.error_mode = error_mode
        self.total_inter_files = 0
        self.total_files = 0
        self.products_with_cve = 0
        self.products_without_cve = 0
        self.cache_dir = cache_dir
        self.merged_files = ["tag"]
        self.score = score

        self.walker = DirWalk(
            pattern=";".join(
                file_path if file_path.endswith(".json") else file_path + "*.json"
                for file_path in self.merge_files
            ),
            yield_files=True,
        ).walk

    def recursive_scan(self, merge_files):
        """Recursive scan all json in a directory/regex path"""
        for intermediate_path in merge_files:
            if os.path.isdir(intermediate_path):
                for filepath in self.walker([intermediate_path]):
                    self.file_stack.append(filepath)
                    yield filepath
                    self.file_stack.pop()
            elif os.path.isfile(intermediate_path) and not os.path.islink(
                intermediate_path
            ):
                self.file_stack.append(intermediate_path)
                yield intermediate_path
                self.file_stack.pop()

    def scan_intermediate_file(self, filename):
        """Reads intermediate json file through filename and verify missing fields"""
        self.logger.info(f"Loading file: {filename}")

        with open(filename) as json_file:
            json_file = json_file.read()
            inter_data = json.loads(json_file)
            if not inter_data or not isinstance(inter_data, dict):
                with ErrorHandler(mode=self.error_mode):
                    raise InvalidJsonError(filename)

            required_fields = set({"metadata", "report"})
            missing_fields = required_fields - set(inter_data.keys())

            if missing_fields == set():
                if isinstance(inter_data["metadata"], dict):

                    missing_fields = set(REQUIRED_INTERMEDIATE_METADATA) - set(
                        inter_data["metadata"].keys()
                    )
                    if missing_fields == set():

                        if isinstance(inter_data["report"], list):
                            self.logger.info(
                                f"Adding data from {os.path.basename(filename)} with timestamp {inter_data['metadata']['timestamp']}"
                            )
                            self.total_inter_files += 1
                            inter_data["metadata"]["severity"] = get_severity_count(
                                inter_data["report"]
                            )
                            return inter_data

            if missing_fields != set():
                with ErrorHandler(mode=self.error_mode):
                    raise MissingFieldsError(f"{missing_fields} are required fields")

            with ErrorHandler(mode=self.error_mode):
                raise InvalidIntermediateJsonError(filename)

    def merge_intermediate(self):
        """Merge valid intermediate dictionaries"""

        for inter_file in self.recursive_scan(self.merge_files):
            # Create a list of intermediate files dictionary
            self.intermediate_cve_data.append(self.scan_intermediate_file(inter_file))

        if self.intermediate_cve_data:
            # sort on basis of timestamp and scans
            self.intermediate_cve_data.sort(
                key=lambda inter: datetime.strptime(
                    inter["metadata"]["timestamp"], "%Y-%m-%d.%H-%M-%S"
                )
            )
            self.all_cve_data = self.remove_intermediate_duplicates()
            merged_cve_scanner = self.get_intermediate_cve_scanner(
                [self.all_cve_data], self.score
            )[0]
            merged_cve_scanner.products_with_cve = self.products_with_cve
            merged_cve_scanner.products_without_cve = self.products_without_cve

            return merged_cve_scanner

        self.logger.error("No valid Intermediate reports found!")
        return {}

    def remove_intermediate_duplicates(self) -> List[Dict[str, str]]:
        """Returns a list of dictionary with same format as cve-bin-tool json output"""

        output = {}
        for inter_data in self.intermediate_cve_data:
            self.products_with_cve += int(inter_data["metadata"]["products_with_cve"])
            self.products_without_cve += int(
                inter_data["metadata"]["products_without_cve"]
            )
            for cve in inter_data["report"]:
                if cve["cve_number"] != "UNKNOWN":
                    if cve["cve_number"] not in output:
                        output[cve["cve_number"]] = cve
                        self.total_files += len(cve["paths"].split(","))
                    else:
                        path_list = output[cve["cve_number"]]["paths"].split(",")
                        self.total_files -= len(path_list)
                        path_list.extend(cve["paths"].split(","))
                        # remove duplicate paths(if any)
                        path_list = list(set(path_list))
                        self.total_files += len(path_list)
                        output[cve["cve_number"]]["path"] = path_list

        return list(output.values())

    @staticmethod
    def get_intermediate_cve_scanner(cve_data_list, score) -> List[CVEScanner]:
        """Returns a list of CVEScanner parsed objects when a list of cve_data json like list is passed"""
        cve_scanner_list = []
        for inter_data in cve_data_list:
            with CVEScanner(score=score) as cve_scanner:
                triage_data: TriageData
                parsed_data: Dict[ProductInfo, TriageData] = {}

                parsed_data = parse_data_from_json(
                    inter_data["report"] if "report" in inter_data else inter_data
                )

                for product_info, triage_data in parsed_data.items():
                    LOGGER.debug(f"{product_info}, {triage_data}")
                    cve_scanner.get_cves(product_info, triage_data)

                cve_scanner_list.append(cve_scanner)

        return cve_scanner_list


def parse_data_from_json(
    json_data: List[Dict[str, str]]
) -> Dict[ProductInfo, TriageData]:
    """Parse CVE JSON dictionary to Dict[ProductInfo, TriageData]"""

    parsed_data: Dict[ProductInfo, TriageData] = defaultdict(dict)

    for row in json_data:
        product_info = ProductInfo(
            row["vendor"].strip(), row["product"].strip(), row["version"].strip()
        )
        parsed_data[product_info][row.get("cve_number", "").strip() or "default"] = {
            "remarks": Remarks(str(row.get("remarks", "")).strip()),
            "comments": row.get("comments", "").strip(),
            "severity": row.get("severity", "").strip(),
        }

        parsed_data[product_info]["paths"] = set(
            map(lambda x: x.strip(), row.get("paths", "").split(","))
        )
    return parsed_data


def get_severity_count(reports: List[Dict[str, str]] = []) -> Dict[str, int]:
    """Returns a list of Severity counts for intermediate report"""
    severity_count = {"LOW": 0, "MEDIUM": 0, "HIGH": 0, "CRITICAL": 0, "UNKNOWN": 0}

    for cve in reports:
        if "severity" in cve and cve["severity"] in severity_count:
            severity_count[cve["severity"]] += 1
        else:
            severity_count["UNKNOWN"] += 1

    return severity_count



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for cups

https://www.cvedetails.com/product/1219/Easy-Software-Products-Cups.html?vendor_id=713

"""

from cve_bin_tool.checkers import Checker


class CupsChecker(Checker):
    CONTAINS_PATTERNS = []
    FILENAME_PATTERNS = [r"cupsd"]
    VERSION_PATTERNS = [r"CUPS v([0-9]+\.[0-9]+\.[0-9]+)"]
    VENDOR_PRODUCT = [("easy_software_products", "cups")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

# pylint: disable=invalid-name

"""
CVE checker for libjpg-turbo

Note that this file is named libjpeg.py instead of libjpeg-turbo.py to avoid an issue
with loading the checker.

References:
https://www.cvedetails.com/vulnerability-list/vendor_id-17075/product_id-40849/Libjpeg-turbo-Libjpeg-turbo.html
"""
from cve_bin_tool.checkers import Checker


class LibjpegChecker(Checker):
    CONTAINS_PATTERNS = [
        r"LIBJPEG",
        r"Caution: quantization tables are too coarse for baseline JPEG",
        r"Invalid JPEG file structure: two SOF markers",
    ]
    FILENAME_PATTERNS = [r"libjpg.so."]
    VERSION_PATTERNS = [
        r"libjpeg-turbo version ([0-9]\.[0-9]\.[0-9])",
        r"LIBJPEGTURBO_([0-9]+\.[0-9]+\.?[0-9]?)",
        r"LIBJPEG_([0-9]+\.[0-9]+\.?[0-9]?)",
    ]
    VENDOR_PRODUCT = [("libjpeg-turbo", "libjpeg-turbo")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

"""
CVE checker for dnsmasq

 https://www.cvedetails.com/vulnerability-list/vendor_id-2959/product_id-5164/Dnsmasq-Dnsmasq.html

"""
from cve_bin_tool.checkers import Checker


class DnsmasqChecker(Checker):
    CONTAINS_PATTERNS = [
        r"Dnsmasq is lightweight, easy to configure DNS forwarder and DHCP server\.\neither in each host or in a central configuration file\."
    ]
    FILENAME_PATTERNS = [r"dnsmasq"]
    VERSION_PATTERNS = [r"dnsmasq-([0-9]+\.[0-9]+)"]
    VENDOR_PRODUCT = [
        ("dnsmasq", "dnsmasq"),
        ("thekelleys", "dnsmasq"),
        ("the_kelleys", "dnsmasq"),
    ]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "libvirt",
        "version": "1.2.3",
        "version_strings": ["LIBVIRT_PRIVATE_1.2.3"],
    }
]
package_test_data = [
    {
        "url": "https://kojipkgs.fedoraproject.org/packages/libvirt/1.2.2/1.fc21/x86_64/",
        "package_name": "libvirt-client-1.2.2-1.fc21.x86_64.rpm",
        "product": "libvirt",
        "version": "1.2.2",
    }
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {"product": "rsyslog", "version": "5.5.6", "version_strings": ["rsyslog 5.5.6"]}
]
package_test_data = [
    {
        "url": "https://kojipkgs.fedoraproject.org/packages/rsyslog/5.5.7/1.fc15/x86_64/",
        "package_name": "rsyslog-5.5.7-1.fc15.x86_64.rpm",
        "product": "rsyslog",
        "version": "5.5.7",
    }
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "libpng",
        "version": "2.0.1",
        "version_strings": [
            "libpng error: %s, offset=%d",
            "libpng version 2.0.1 -",
            "libpng version ",
        ],
    }
]
package_test_data = [
    {
        "url": "https://archives.fedoraproject.org/pub/archive/fedora/linux/releases/30/Everything/x86_64/os/Packages/l/",
        "package_name": "libpng-1.6.36-1.fc30.x86_64.rpm",
        "product": "libpng",
        "version": "1.6.36",
    },
    {
        "url": "http://mirror.centos.org/centos/7/os/x86_64/Packages/",
        "package_name": "libpng-1.5.13-8.el7.x86_64.rpm",
        "product": "libpng",
        "version": "1.5.13",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {"product": "tcpdump", "version": "4.9.0", "version_strings": ["tcpdump-4.9.0"]}
]
package_test_data = [
    {
        "url": "https://kojipkgs.fedoraproject.org/packages/tcpdump/4.9.1/1.fc27/x86_64/",
        "package_name": "tcpdump-4.9.1-1.fc27.x86_64.rpm",
        "product": "tcpdump",
        "version": "4.9.1",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for lighttpd

References:
https://www.cvedetails.com/vulnerability-list/vendor_id-2713/product_id-4762/Lighttpd-Lighttpd.html

RSS feed: http://www.cvedetails.com/vulnerability-feed.php?vendor_id=2713&product_id=4762&version_id=0&orderby=3&cvssscoremin=0

"""
from cve_bin_tool.checkers import Checker


class LighttpdChecker(Checker):
    CONTAINS_PATTERNS = [
        r"Invalid fds at startup with lighttpd",
        r"lighttpd will fail to start up",
    ]
    FILENAME_PATTERNS = [r"lighttpd"]
    VERSION_PATTERNS = [r"lighttpd/([0-9]+\.[0-9]+\.[0-9]+)"]
    VENDOR_PRODUCT = [("lighttpd", "lighttpd")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "libsoup",
        "version": "2.70.0",
        "version_strings": ["libsoup/2.70.0"],
    },
    {
        "product": "libsoup",
        "version": "2.62.3",
        "version_strings": ["libsoup/2.62.3"],
    },
]
package_test_data = [
    {
        "url": "https://download-ib01.fedoraproject.org/pub/fedora/linux/releases/32/Everything/aarch64/os/Packages/l/",
        "package_name": "libsoup-2.70.0-1.fc32.aarch64.rpm",
        "product": "libsoup",
        "version": "2.70.0",
    },
    {
        "url": "http://mirror.centos.org/centos/8/BaseOS/aarch64/os/Packages/",
        "package_name": "libsoup-2.62.3-2.el8.aarch64.rpm",
        "product": "libsoup",
        "version": "2.62.3",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

"""
CVE checker for libsoup

https://www.cvedetails.com/product/21096/Gnome-Libsoup.html?vendor_id=283
"""
from cve_bin_tool.checkers import Checker


class LibsoupChecker(Checker):
    CONTAINS_PATTERNS = []
    FILENAME_PATTERNS = [r"libsoup"]
    VERSION_PATTERNS = [r"libsoup/([0-9]+\.[0-9]+\.[0-9]+)"]
    VENDOR_PRODUCT = [
        ("gnome", "libsoup"),
        ("joe_shaw", "libsoup"),
        ("libsoup", "libsoup"),
    ]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

import csv
import json
import os
from collections import defaultdict
from logging import Logger
from typing import Any, DefaultDict, Dict, Iterable, Set, Union

from cve_bin_tool.error_handler import (
    ErrorHandler,
    ErrorMode,
    InvalidCsvError,
    InvalidJsonError,
    MissingFieldsError,
)
from cve_bin_tool.log import LOGGER
from cve_bin_tool.util import ProductInfo, Remarks

# TriageData is dictionary of cve_number mapped to dictionary of remarks, comments and custom severity
TriageData = Dict[str, Union[Dict[str, Any], Set[str]]]


class InputEngine:
    # parsed_data is a dictionary of vendor, product, version mapped to TriageData
    parsed_data: DefaultDict[ProductInfo, TriageData]

    def __init__(
        self, filename: str, logger: Logger = None, error_mode=ErrorMode.TruncTrace
    ):
        self.filename = os.path.abspath(filename)
        self.logger = logger or LOGGER.getChild(self.__class__.__name__)
        self.error_mode = error_mode
        self.parsed_data = defaultdict(dict)

    def parse_input(self) -> DefaultDict[ProductInfo, TriageData]:
        if not os.path.isfile(self.filename):
            with ErrorHandler(mode=self.error_mode):
                raise FileNotFoundError(self.filename)
        if self.filename.endswith(".csv"):
            self.input_csv()
        elif self.filename.endswith(".json"):
            self.input_json()
        return self.parsed_data

    def input_csv(self) -> None:
        with open(self.filename) as csv_file:
            csvdata = csv.DictReader(csv_file)
            if csvdata is None or csvdata.fieldnames is None:
                with ErrorHandler(mode=self.error_mode):
                    raise InvalidCsvError(self.filename)

            self.parse_data(set(csvdata.fieldnames), csvdata)

    def input_json(self) -> None:
        with open(self.filename) as json_file:
            json_data = json.load(json_file)
            if not json_data or not isinstance(json_data, list):
                with ErrorHandler(mode=self.error_mode):
                    raise InvalidJsonError(self.filename)

            self.parse_data(set(json_data[0].keys()), json_data)

    def parse_data(self, fields: Set[str], data: Iterable) -> None:
        required_fields = {"vendor", "product", "version"}
        missing_fields = required_fields - fields
        if missing_fields != set():
            with ErrorHandler(mode=self.error_mode):
                raise MissingFieldsError(f"{missing_fields} are required fields")

        for row in data:
            product_info = ProductInfo(
                row["vendor"].strip(), row["product"].strip(), row["version"].strip()
            )
            self.parsed_data[product_info][
                row.get("cve_number", "").strip() or "default"
            ] = {
                "remarks": Remarks(str(row.get("remarks", "")).strip()),
                "comments": row.get("comments", "").strip(),
                "severity": row.get("severity", "").strip(),
            }
            self.parsed_data[product_info]["paths"] = set(
                map(lambda x: x.strip(), row.get("paths", "").split(","))
            )



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for libnss

References:
http://www.cvedetails.com/vulnerability-list/vendor_id-452/product_id-4052/Mozilla-Network-Security-Services.html

RSS feed: http://www.cvedetails.com/vulnerability-feed.php?vendor_id=452&product_id=4052&version_id=&orderby=3&cvssscoremin=0
"""
from cve_bin_tool.checkers import Checker


class LibnssChecker(Checker):
    CONTAINS_PATTERNS = [
        r"Client authentication failed: private key in key database does not match public key in certificate database.",
        r"The next protocol negotiation extension was enabled, but the callback was cleared prior to being needed.",
        r"# SSL/TLS secrets log file, generated by NSS",
        r"Header: NSS ",
        r"Version: NSS ",
    ]
    FILENAME_PATTERNS = [r"libnss.so.", r"libssl3.so"]
    VERSION_PATTERNS = [
        r"\$Header: NSS ([01234]+\.[0-9]+) ",
        r"\$Header: NSS ([01234]+\.[0-9]+\.[0-9]+) ",
        r"\$Header: NSS ([01234]+\.[0-9]+\.[0-9]+)\.[0-9]+ ",
        r"Version: NSS ([01234]+\.[0-9]+) Basic",
        r"Version: NSS ([01234]+\.[0-9]+\.[0-9]+) Basic",
        r"Version: NSS ([01234]+\.[0-9]+\.[0-9]+)\.[0-9]+ Basic",
        r"Version: NSS (\d+(?:\.\d+)+)",
        r"\@\(#\)NSS ([01234]+\.[0-9]+) Basic",
        r"\@\(#\)NSS ([01234]+\.[0-9]+\.[0-9]+) Basic",
        r"\@\(#\)NSS ([01234]+\.[0-9]+\.[0-9]+)\.[0-9]+ Basic",
    ]
    VENDOR_PRODUCT = [("mozilla", "network_security_services")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

"""
CVE checker for GnuTLS
References:
https://www.cvedetails.com/vulnerability-list/vendor_id-72/product_id-4433/GNU-Gnutls.html
"""
from cve_bin_tool.checkers import Checker


class GnutlsChecker(Checker):
    # FIXME: fix contains pattern
    CONTAINS_PATTERNS = []
    FILENAME_PATTERNS = [
        "gnutls-cli",
        "libgnutls.so",
        "libgnutls-dane.so",
        "gnutls-serv",
    ]
    VERSION_PATTERNS = [
        r"gnutls-cli ([0-9]+\.[0-9]+\.[0-9]+)",
        r"gnutls-serv ([0-9]+\.[0-9]+\.[0-9]+)",
    ]
    VENDOR_PRODUCT = [("gnutls", "gnutls"), ("gnu", "gnutls")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "enscript",
        "version": "1.6.6",
        "version_strings": [
            "set the PostScript language level that enscript",
            "or set the environment variable `ENSCRIPT_LIBRARY' to point to your library directory.",
            "GNU Enscript 1.6.6",
        ],
    },
    {
        "product": "enscript",
        "version": "1.6.5",
        "version_strings": [
            "set the PostScript language level that enscript",
            "or set the environment variable `ENSCRIPT_LIBRARY' to point to your library directory.",
            "GNU Enscript 1.6.5",
        ],
    },
]
package_test_data = [
    {
        "url": "http://mirror.centos.org/altarch/7/os/aarch64/Packages/",
        "package_name": "enscript-1.6.6-7.el7.aarch64.rpm",
        "product": "enscript",
        "version": "1.6.6",
    },
    {
        "url": "http://archive.ubuntu.com/ubuntu/pool/universe/e/enscript/",
        "package_name": "enscript_1.6.5.90-3_amd64.deb",
        "product": "enscript",
        "version": "1.6.5",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

import textwrap
from collections import defaultdict
from datetime import datetime
from typing import DefaultDict, Dict, List

from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from ..cve_scanner import CVEData
from ..input_engine import Remarks
from ..linkify import linkify_cve
from ..theme import cve_theme
from ..util import ProductInfo


def output_console(
    all_cve_data: Dict[ProductInfo, CVEData],
    time_of_last_update,
    console=Console(theme=cve_theme),
):
    """Output list of CVEs in a tabular format with color support"""

    console._width = 120
    now = datetime.now().strftime("%Y-%m-%d  %H:%M:%S")
    time_of_last_update = time_of_last_update.strftime("%Y-%m-%d  %H:%M:%S")

    console.print(
        Markdown(
            textwrap.dedent(
                f"""
                # CVE BINARY TOOL
                - cve-bin-tool Report Generated: {now}
                - Time of last update of CVE Data: {time_of_last_update}
                """
            )
        )
    )

    remarks_colors = {
        Remarks.Mitigated: "green",
        Remarks.Confirmed: "red",
        Remarks.NewFound: "blue",
        Remarks.Unexplored: "yellow",
        Remarks.Ignored: "white",
    }

    cve_by_remarks: DefaultDict[Remarks, List[Dict[str, str]]] = defaultdict(list)
    # group cve_data by its remarks
    for product_info, cve_data in all_cve_data.items():
        for cve in cve_data["cves"]:
            cve_by_remarks[cve.remarks].append(
                {
                    "vendor": product_info.vendor,
                    "product": product_info.product,
                    "version": product_info.version,
                    "cve_number": cve.cve_number,
                    "severity": cve.severity,
                    "score": cve.score,
                    "cvss_version": cve.cvss_version,
                }
            )

    for remarks in sorted(cve_by_remarks):
        color = remarks_colors[remarks]
        console.print(Panel(f"[{color}] {remarks.name} CVEs [/{color}]", expand=False))
        # table instance
        table = Table()

        # Add Head Columns to the Table
        table.add_column("Vendor")
        table.add_column("Product")
        table.add_column("Version")
        table.add_column("CVE Number")
        table.add_column("Severity")
        table.add_column("Score (CVSS Version)")
        # table.add_column("CVSS Version")

        for cve_data in cve_by_remarks[remarks]:
            color = cve_data["severity"].lower()
            table.add_row(
                Text.styled(cve_data["vendor"], color),
                Text.styled(cve_data["product"], color),
                Text.styled(cve_data["version"], color),
                linkify_cve(Text.styled(cve_data["cve_number"], color)),
                Text.styled(cve_data["severity"], color),
                Text.styled(
                    str(cve_data["score"])
                    + " (v"
                    + str(cve_data["cvss_version"])
                    + ")",
                    color,
                ),
            )
        # Print the table to the console
        console.print(table)
        for cve_data in cve_by_remarks[remarks]:
            if "*" in cve_data["vendor"]:
                console.print("* vendors guessed by the tool")
                break



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "kerberos",
        "version": "5-1.15.1",
        "version_strings": [
            "An unknown option was passed in to kerberos",
            "CLIENT kerberos 5-1.15.1",
            "KRB5_BRAND: ",
        ],
    },
    {
        "product": "kerberos",
        "version": "5-1.15.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1",
        "version_strings": [
            "An unknown option was passed in to kerberos",
            "CLIENT kerberos 5-1.15.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1.1",
            "KRB5_BRAND: ",
        ],
    },
]
package_test_data = {}



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for libdb (berkeley db)
CVE list: https://www.cvedetails.com/vulnerability-list/vendor_id-93/product_id-32070/Oracle-Berkeley-Db.html
"""
from cve_bin_tool.checkers import Checker


class LibdbChecker(Checker):
    CONTAINS_PATTERNS = [
        "BDB1568 Berkeley DB library does not support DB_REGISTER on this system",
        "BDB1507 Thread died in Berkeley DB library",
        "Berkeley DB ",
    ]
    FILENAME_PATTERNS = [r"libdb-"]
    VERSION_PATTERNS = [
        r"Berkeley DB ([0-9]+\.[0-9]+\.[0-9]+):",  # short version as backup. we mostly want the long below.
        r"Berkeley DB .+, library version ([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+):",
    ]
    VENDOR_PRODUCT = [("oracle", "berkeley_db")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "expat",
        "version": "2.0.1",
        "version_strings": [
            "expat_2.0.1 ",
            "requested feature requires XML_DTD support in Expat",
        ],
    }
]
package_test_data = [
    {
        "url": "http://mirror.centos.org/centos/7/os/x86_64/Packages/",
        "package_name": "expat-2.1.0-12.el7.x86_64.rpm",
        "product": "expat",
        "version": "2.1.0",
    },
    {
        "url": "https://kojipkgs.fedoraproject.org/packages/expat/2.2.1/1.fc24/x86_64/",
        "package_name": "expat-2.2.1-1.fc24.x86_64.rpm",
        "product": "expat",
        "version": "2.2.1",
    },
    {
        "url": "http://http.us.debian.org/debian/pool/main/e/expat/",
        "package_name": "libexpat1_2.2.0-2+deb9u3_amd64.deb",
        "product": "expat",
        "version": "2.2.0",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

"""
CVE-bin-tool util tests
"""
import inspect
from typing import DefaultDict

from cve_bin_tool.cve_scanner import CVEScanner
from cve_bin_tool.util import CVEData, ProductInfo, inpath


class TestUtil:
    """Test the util functions"""

    def test_inpath(self):
        """Test the check to see if a command line utility is installed
        and in path before we try to run it."""
        assert inpath("python")

    def test_not_inpath(self):
        assert not inpath("cve_bin_tool_test_for_not_in_path")


class TestSignature:
    """Tests signature of critical class and functions"""

    def test_cve_scanner(self):
        sig = inspect.signature(CVEScanner.get_cves)
        expected_args = {"product_info", "triage_data", "self"}
        assert (
            set(sig.parameters) - expected_args == set()
        ), "Parameters of get_cves has been changed. Make sure it isn't breaking InputEngine!"

        instance_attrs = vars(CVEScanner)["__annotations__"]
        assert (
            instance_attrs["all_cve_data"] == DefaultDict[ProductInfo, CVEData]
        ), "Type of all_cve_data has been changed. Make sure it isn't breaking OutputEngine!"



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "binutils",
        "version": "2.31.1",
        "version_strings": [
            "Using the --size-sort and --undefined-only options together",
            "libbfd-2.31.1-system.so",
            "Auxiliary filter for shared object symbol table",
        ],
    }
]
package_test_data = [
    {
        "url": "http://security.ubuntu.com/ubuntu/pool/main/b/binutils/",
        "package_name": "binutils_2.26.1-1ubuntu1~16.04.8_amd64.deb",
        "product": "binutils",
        "version": "2.26.1",
    },
    {
        "url": "http://mirror.centos.org/centos/7/os/x86_64/Packages/",
        "package_name": "binutils-2.27-44.base.el7.x86_64.rpm",
        "product": "binutils",
        "version": "2.27",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {"product": "gimp", "version": "2.6.10", "version_strings": ["GIMP 2.6.10"]}
]
package_test_data = [
    {
        "url": "https://kojipkgs.fedoraproject.org/packages/gimp/2.6.12/1.fc15/x86_64/",
        "package_name": "gimp-2.6.12-1.fc15.x86_64.rpm",
        "product": "gimp",
        "version": "2.6.12",
    },
    {
        "url": "http://ftp.osuosl.org/pub/ubuntu/pool/universe/g/gimp/",
        "package_name": "gimp_2.8.22-1_amd64.deb",
        "product": "gimp",
        "version": "2.8.22",
    },
    {
        "url": "http://mirror.centos.org/centos/7/os/x86_64/Packages/",
        "package_name": "gimp-2.8.22-1.el7.x86_64.rpm",
        "product": "gimp",
        "version": "2.8.22",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "network_security_services",
        "version": "3.37.1",
        "version_strings": [
            "$Header: NSS 3.37.1 ",
            "Version: NSS 3.37.1 Basic",
            "@#NSS 3.37.1 Basic",
            "# SSL/TLS secrets log file, generated by NSS",
        ],
    }
]
package_test_data = [
    {
        "url": "http://mirror.centos.org/altarch/7/os/aarch64/Packages/",
        "package_name": "nss-3.44.0-7.el7_7.aarch64.rpm",
        "product": "network_security_services",
        "version": "3.44",
    },
    {
        "url": "https://kojipkgs.fedoraproject.org/packages/nss/3.37.3/3.fc29/x86_64/",
        "package_name": "nss-3.37.3-3.fc29.x86_64.rpm",
        "product": "network_security_services",
        "version": "3.37.3",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "logrotate",
        "version": "3.18.0",
        "version_strings": [
            "logrotate 3.18.0 - Copyright (C) 1995-2001 Red Hat, Inc.",
            "WARNING: logrotate in debug mode does nothing except printing debug messages!  Consider using verbose mode (-v) instead if this is not what you want.",
            'Set "su" directive in config file to tell logrotate which user/group should be used for rotation',
        ],
    },
    {
        "product": "logrotate",
        "version": "3.14.0",
        "version_strings": [
            "logrotate 3.14.0 - Copyright (C) 1995-2001 Red Hat, Inc.",
            "WARNING: logrotate in debug mode does nothing except printing debug messages!  Consider using verbose mode (-v) instead if this is not what you want.",
            'Set "su" directive in config file to tell logrotate which user/group should be used for rotation',
        ],
    },
]
package_test_data = [
    {
        "url": "https://download-ib01.fedoraproject.org/pub/fedora/linux/releases/34/Everything/aarch64/os/Packages/l/",
        "package_name": "logrotate-3.18.0-2.fc34.aarch64.rpm",
        "product": "logrotate",
        "version": "3.18.0",
    },
    {
        "url": "http://mirror.centos.org/centos/8/BaseOS/aarch64/os/Packages/",
        "package_name": "logrotate-3.14.0-4.el8.aarch64.rpm",
        "product": "logrotate",
        "version": "3.14.0",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for mariadb CLI

References:
https://github.com/MariaDB/server
https://www.cvedetails.com/vulnerability-list/vendor_id-12010/product_id-22503/Mariadb-Mariadb.html


"""
from cve_bin_tool.checkers import Checker


class MariadbChecker(Checker):

    CONTAINS_PATTERNS = [
        r"Oracle, MariaDB Corporation Ab and others.",
        r"General information about MariaDB can be found at\nhttp://mariadb.org",
        r"Welcome to the MariaDB monitor.",
        r"MariaDB virtual IO plugin for socket communication",
    ]
    FILENAME_PATTERNS = [
        r"mariadb",
        r"mariadb-client",
        r"mariadb_config",
        r"mariadb-client-core",
        r"mariadb-test",
        r"mariadb-test-data",
        r"odbc-mariadb",
        r"libmariadb-dev",
        r"libmariadb-dev-compat",
        r"libmariadb-java",
        r"libmariadbclient-dev",
        r"mariadb-backup",
        r"mariadb-plugin",
    ]
    VERSION_PATTERNS = [r"([0-9]+\.[0-9]+\.[0-9]+)-MariaDB"]
    VENDOR_PRODUCT = [("mariadb", "mariadb")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "zlib",
        "version": "1.2.2",
        "version_strings": [
            "deflate 1.2.2 Copyright 1995-2005 Jean-loup Gailly",
            "too many length or distance symbols",
        ],
    }
]
package_test_data = [
    {
        "url": "https://kojipkgs.fedoraproject.org/packages/zlib/1.2.11/23.fc33/aarch64/",
        "package_name": "zlib-1.2.11-23.fc33.aarch64.rpm",
        "product": "zlib",
        "version": "1.2.11",
    },
    {
        "url": "http://mirror.centos.org/centos/7/os/x86_64/Packages/",
        "package_name": "zlib-1.2.7-18.el7.x86_64.rpm",
        "product": "zlib",
        "version": "1.2.7",
    },
    {
        "url": "http://archive.ubuntu.com/ubuntu/pool/main/z/zlib/",
        "package_name": "zlib1g_1.2.8.dfsg-2ubuntu4_amd64.deb",
        "product": "zlib",
        "version": "1.2.8",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for bzip2

https://www.cvedetails.com/vulnerability-list/vendor_id-1198/product_id-2068/Bzip-Bzip2.html

"""
from cve_bin_tool.checkers import Checker


class Bzip2Checker(Checker):
    # FIXME: fix contains pattern
    CONTAINS_PATTERNS = []
    FILENAME_PATTERNS = [r"bzip2"]
    VERSION_PATTERNS = [r"bzip2-([0-9]+\.[0-9]+\.[0-9]+)"]
    VENDOR_PRODUCT = [("bzip", "bzip2")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "ncurses",
        "version": "6.1",
        "version_strings": ["_nc_first_name", "ncurses 6.1"],
    }
]
package_test_data = [
    {
        "url": "https://kojipkgs.fedoraproject.org/packages/ncurses/6.2/1.20200222.fc33/x86_64/",
        "package_name": "ncurses-6.2-1.20200222.fc33.x86_64.rpm",
        "product": "ncurses",
        "version": "6.2",
    },
    {
        "url": "https://mirrors.edge.kernel.org/centos/7/os/x86_64/Packages/",
        "package_name": "ncurses-libs-5.9-14.20130511.el7_4.x86_64.rpm",
        "product": "ncurses",
        "version": "5.9",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "webkitgtk",
        "version": "1.5.12",
        "version_strings": ["webkitgtk-1.5.12"],
    }
]
package_test_data = [
    {
        "url": "http://mirror.centos.org/centos/7/os/x86_64/Packages/",
        "package_name": "webkitgtk3-2.4.11-2.el7.x86_64.rpm",
        "product": "webkitgtk",
        "version": "2.4.11",
    },
    {
        "url": "https://kojipkgs.fedoraproject.org/packages/webkitgtk/2.0.1/1.fc19/x86_64/",
        "package_name": "webkitgtk-2.0.1-1.fc19.x86_64.rpm",
        "product": "webkitgtk",
        "version": "2.0.1",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "bolt",
        "version": "0.9",
        "version_strings": ["bolt 0.9", "bolt 0.9 starting up."],
    },
    {
        "product": "bolt",
        "version": "0.9.1",
        "version_strings": ["bolt 0.9.1", "bolt 0.9.1 starting up."],
    },
]
package_test_data = [
    {
        "url": "http://archive.ubuntu.com/ubuntu/pool/main/b/bolt/",
        "package_name": "bolt_0.9.1-1_amd64.deb",
        "product": "bolt",
        "version": "0.9.1",
    },
    {
        "url": "https://download-ib01.fedoraproject.org/pub/fedora/linux/releases/33/Everything/aarch64/os/Packages/b/",
        "package_name": "bolt-0.9-3.fc33.aarch64.rpm",
        "product": "bolt",
        "version": "0.9",
    },
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for freeradius

https://www.cvedetails.com/product/1805/Freeradius-Freeradius.html?vendor_id=1039

"""
from cve_bin_tool.checkers import Checker


class FreeradiusChecker(Checker):
    CONTAINS_PATTERNS = []
    FILENAME_PATTERNS = [r"radiusd"]
    VERSION_PATTERNS = [r"FreeRADIUS Version ([0-9]+\.[0-9]+\.[0-9]+)"]
    VENDOR_PRODUCT = [("freeradius", "freeradius")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for hostapd

https://www.cvedetails.com/product/22495/W1.fi-Hostapd.html?vendor_id=12005

"""
from cve_bin_tool.checkers import Checker


class HostapdChecker(Checker):
    # FIXME: fix contains pattern
    CONTAINS_PATTERNS = []
    FILENAME_PATTERNS = [r"hostapd"]
    VERSION_PATTERNS = [r"hostapd v([0-9]+\.[0-9]+)"]
    VENDOR_PRODUCT = [("w1.fi", "hostapd")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

mapping_test_data = [
    {
        "product": "wpa_supplicant",
        "version": "2.8",
        "version_strings": ["wpa_supplicant v2.8"],
    }
]
package_test_data = [
    {
        "url": "http://mirror.centos.org/centos/7/os/x86_64/Packages/",
        "package_name": "wpa_supplicant-2.6-12.el7.x86_64.rpm",
        "product": "wpa_supplicant",
        "version": "2.6",
    }
]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later


"""
CVE checker for libvirt

https://www.cvedetails.com/product/15743/Libvirt-Libvirt.html?vendor_id=8917

"""
from cve_bin_tool.checkers import Checker


class LibvirtChecker(Checker):
    CONTAINS_PATTERNS = []
    FILENAME_PATTERNS = [
        r"libvirtd",
        r"libvirt.so",
    ]
    VERSION_PATTERNS = [r"LIBVIRT_PRIVATE_([0-9]+\.[0-9]+\.[0-9]+)"]
    VENDOR_PRODUCT = [("libvirt", "libvirt")]



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

# pylint: disable=invalid-name
"""
CVE checker for systemd

https://www.cvedetails.com/product/38088/Freedesktop-Systemd.html?vendor_id=7971
"""

from cve_bin_tool.checkers import Checker

from ..util import regex_find


class SystemdChecker(Checker):
    CONTAINS_PATTERNS = [
        r"sd_bus_error_copy",
        r"sd_bus_error_is_set",
        r"sd_bus_error_add_map",
    ]
    FILENAME_PATTERNS = [r"libsystemd.so."]
    VERSION_PATTERNS = [
        r"LIBSYSTEMD_([0-4]+[0-9]+[0-9]+)",
        r"^systemd (\d{2,4})$",
        r"libsystemd-shared-([0-9]+)\.so",  # patterns like this aren't ideal
        r"systemd-[a-z]+-([0-9]+)\.so",  # patterns like this aren't ideal
        r"udev-([0-9]+)\.so",  # patterns like this aren't ideal
        r"systemd v([0-9]+).* running in ",
        r"pam_systemd.so-([0-9]+)\.",
    ]
    VENDOR_PRODUCT = [("freedesktop", "systemd")]

    """
    Using filenames (containing patterns like '.so' etc.) in the binaries as VERSION_PATTERNS aren't ideal.
    The reason behind this is that these might depend on who packages the file (like it 
    might work on fedora but not on ubuntu)
    """

    def get_version(self, lines, filename):
        def guess_contains_systemd(lines):
            """Tries to determine if a file includes systemd"""
            for line in lines:
                if "sd_bus_error_copy" in line:
                    return 1
                if "sd_bus_error_is_set" in line:
                    return 1
                if "sd_bus_error_add_map" in line:
                    return 1
            return 0

        version_info = {}

        if "libsystemd.so." in filename:
            version_info["is_or_contains"] = "is"

        elif guess_contains_systemd(lines):
            version_info["is_or_contains"] = "contains"

        if "is_or_contains" in version_info:
            version_info["modulename"] = "systemd"
            version_info["version"] = regex_find(
                sorted(lines, reverse=True), self.VERSION_PATTERNS
            )

        return version_info
